import numpy as np


def logistic_map(n_timesteps: int, r: float=3.9, x0: float=0.5)->np.ndarray:
    """Logistic map timeseries:
    .. math::
        x(n+1) = rx(n)(1-x(n))

    Args:
        n_timesteps (float): 
        r (float, optional): r parameter of the logistic map equation. Value should be between
                              3.6 and 3.9 to show intersting chaotic behaviour. Values between 3.45 and 3.56
                              expose a periodic behaviour beteen a increasing number of values. Defaults to 3.9.
        x0 (float, optional): Initial condition of the timeseries. Should be in ]0; 1[. Defaults to 0.5.

    Raises:
        ValueError: if x0 is outside ]0; 1[.
        ValueError: if r is not strictly positive.

    Returns:
        np.ndarray: Logistic map timeseries.
    """
    if r>0 and 0<x0<1:
        X = np.zeros(n_timesteps)
        X[0] = x0

        for i in range(1, n_timesteps):
            X[i] = r * X[i-1] * (1-X[i-1])

        return X
    elif r<=0:
        raise ValueError("r should be positive.")
    else:
        raise ValueError("Initial condition x0 should be in ]0;1[.")


from . import mat_gen
from . import observables
from . import regression_models
from . import activationsfunc
from . import hyper
from .esn import ESN
from .esn_online import ESNOnline
from .utils import load

__version__ = "0.2.1"



import numpy as np
from scipy import linalg

def get_spectral_radius(W):
    """Given a squarred weight matrix (e.g. the recurrent reservoir matrix),
    returns the spectral radius"""
    return max(abs(linalg.eig(W)[0]))

def compute_error_NRMSE(teacher_signal, predicted_signal, verbose=False):
    """ Computes Normalized Root-Mean-Squarred Error between a teacher signal and a predicted signal
    Return the errors in this order: nmrse mean, nmrse max-min, rmse, mse.
    By default, only NMRSE mean should be considered as a general measure to be
    compared for different datasets.

    For more information, see:
    - Mean Squared Error https://en.wikipedia.org/wiki/Mean_squared_error
    - Root Mean Squared Error https://en.wikipedia.org/wiki/Root-mean-square_deviation for more info
    """
    errorLen = len(predicted_signal[:])
    mse = np.mean((teacher_signal - predicted_signal)**2)
    rmse = np.sqrt(mse)
    nmrse_mean = abs(rmse / np.mean(predicted_signal[:])) # Normalised RMSE (based on mean)
    nmrse_maxmin = rmse / abs(np.max(predicted_signal[:]) - np.min(predicted_signal[:])) # Normalised RMSE (based on max - min)
    if verbose:
        print("Errors computed over %d time steps" % (errorLen))
        print("\nMean Squared error (MSE):\t\t%.4e" % (mse) )
        print("Root Mean Squared error (RMSE):\t\t%.4e\n" % rmse )
        print("Normalized RMSE (based on mean):\t%.4e" % (nmrse_mean) )
        print("Normalized RMSE (based on max - min):\t%.4e" % (nmrse_maxmin) )
    return nmrse_mean, nmrse_maxmin, rmse, mse



from typing import Tuple 

import numpy as np

from scipy.integrate import solve_ivp


def multiscroll(n_timesteps: int, 
            state0: list=[-0.1, 0.5, -0.6], 
            a: float=40.0, 
            b: float=3.0, 
            c: float=28.0, 
            h: float=0.01) -> Tuple[np.ndarray, np.ndarray]:
    """Double scroll attractor timeseries, a particular case of multiscroll attractor.
    .. math::
        \dv{x}{t} = a(y - x)
    .. math::
        \dv{y}{t} = (c - a)x - xz + cy
    .. math::
        \dv{z}{t} = xy - bz

    Args:
        n_timesteps (int): Number of timesteps to generate.
        state0 (list, optional): Initial conditions of the system. Defaults to [-0.1, 0.5, -0.6].
        a (float, optional): a parameter of the system. Defaults to 40.0.
        b (float, optional): b parameter of the system. Defaults to 3.0.
        c (float, optional): c parameter of the system. Defaults to 28.0.
        h (float, optional): Control the time delta between two timesteps. Defaults to 0.01.

    Returns:
        Tuple[np.ndarray, np.ndarray]: Generated timeseries and timesteps.
    """

    def multiscroll_diff(t, state):
        x, y, z = state
        dx = a * (y - x)
        dy = (c - a)*x - x*z + c*y
        dz = x*y - b*z
        return dx, dy, dz

    t = np.arange(0.0, n_timesteps * h, h)
    
    sol = solve_ivp(multiscroll_diff,
                    y0=state0,
                    t_span=(0.0, n_timesteps*h),
                    dense_output=True)

    return sol.sol(t).T, t


from typing import Tuple

import collections

import numpy as np


DEFAULT_SEED = 5555


def mackey_glass(n_timesteps: int, 
                 tau: int=17, 
                 a: float=0.2, 
                 b: float=0.1, 
                 n: int=10, 
                 x0: float=1.2, 
                 h: float=1.0, 
                 seed: int=None) -> Tuple[np.ndarray, np.ndarray]:
    """Mackey-Glass timeseries, computed from the Mackey-Glass delay differential equation:
    .. math:: 
        \dv{x}{t} = \frac{ax(t-\tau)}{1+x(t-\tau)^n} - bx(t)
    
    Args:
        n_timesteps (int): Number of timesteps to compute.
        tau (int, optional): Time delay of Mackey-Glass equation. Defaults to 17.
        a (float, optional): Parameter a of equation.. Defaults to 0.2.
        b (float, optional): Parameter b of equation. Defaults to 0.1.
        n (int, optional): Parameter n of equation. Defaults to 10.
        x0 (float, optional): Initial value of the timeseries. Defaults to 1.2.
        h (float, optional): Time delta for the Runge-Kuta method. Can be assimilated
                             to the number of discrete point computed per timestep. Defaults to 1.0.
        seed (int, optional): Random state seed for reproducibility. Defaults to None.

    Returns:
        np.ndarray, np.ndarray: Computed timeseries and corresponding timesteps.
    """
    
    if seed is not None:
        rs = np.random.RandomState(seed)
    else:
        rs = np.random.RandomState(DEFAULT_SEED)
    
    
    time = 0
    history_length = int(np.floor(tau/h))
    history = collections.deque(x0 * np.ones(history_length) + \
                                0.2 * (rs.rand(history_length) - 0.5))
    xt = x0

    X = np.zeros(n_timesteps)
    T = np.zeros(n_timesteps)

    for i in range(0, n_timesteps):
        X[i] = xt
        
        if tau == 0:
            xtau = 0.0
        else:
            xtau = history.popleft()
            history.append(xt)

        xth = _mc_rk4(xt, xtau, a=a, b=b, n=n)

        time = time + h
        T[i] = time
        xt = xth
    
    return X, T


def _mc_eq(xt, xtau, a=0.2, b=0.1, n=10):
    """
    Mackey-Glass time delay diffential equation, at values x(t) and x(t-tau):
    .. math:: \dv{x}{t} = \frac{ax(t-\tau)}{1+x(t-tau)^n} - bx(t)
    
    Arguments:
        xt {float} -- Value x(t) of the Mackey-Glass timeseries.
        xtau {float} -- Value x(t - tau) of the Mackey-Glass timeseries.
    Keywork arguments:
        a {float} -- Parameter a of the Mackey-Glass delay differential equation.
        b {float} -- Parameter b of the Mackey-Glass delay differential equation.
        n {float} -- Parameter n of the Mackey-Glass delay differential equation.
    Returns:
        {float} -- The value of the differential at a point t.
    """
    return -b*xt + a*xtau / (1+xtau**n)

    
def _mc_rk4(xt, xtau, a, b, n, h=1.0):
    """
    Runge-Kuta method (RK4) for Mackey-Glass timeseries discretization.
    
    Arguments:
        xt {float} -- Value x(t) of the Mackey-Glass timeseries.
        xtau {float} -- Value x(t - tau) of the Mackey-Glass timeseries.
    Keyword arguments:
        h {float} -- Time delta for the Runge-Kuta method. Can be assimilated
                     to the number of discrete point computed per timestep.
    Returns:
        {float} -- Runge-Kuta approximation of the timeseries at point t.
    """
    k1 = h * _mc_eq(xt, xtau, a, b, n)
    k2 = h * _mc_eq(xt + 0.5*k1, xtau, a, b, n)
    k3 = h * _mc_eq(xt + 0.5*k2, xtau, a, b, n)
    k4 = h * _mc_eq(xt + k3, xtau, a, b, n)
    
    return xt + k1/6 + k2/3 + k3/3 + k4/6


"""Quick tools for weight matrices initialization.

This module provides simples tools for reservoir internal weights
and input/feedback weights initialization. Spectral radius of the
internal weights, input scaling and sparsity are fully parametrizable.

Because most of the architectures developped in *reservoir computing*
involve sparsely-connected neuronal units, the prefered format for all
generated matrices is a `scipy.sparse` format (in most cases *csr*).
Sparse arrays allow fast computations and compact representations of
weights matrices, and remains easily readable. They can be parsed back to
simple Numpy arrays just by calling their `toarray()`̀ method.

All functions can take as paramater a `numpy.random.RandomState`
instance, or a seed number, to ensure reproducibility. Both distribution
of weights and distribution of non-zero connections are controled with the
seed.

Example
-------

Here, we generate a 1000 units reservoir `W` with a spectral radius of 0.5,
connected to 5 inputs by the `W_in` matrix, with an input scaling of 0.9.

.. code-block:: python

    from reservoirpy.mat_gen import fast_spectral_initialization
    from reservoirpy.mat_gen import generate_input_weights
    W = fast_spectral_initialization(1000, spectral_radius=0.5)
    Win = generate_input_weights(1000, 5, input_scaling=0.9)
"""
from typing import Union

import numpy as np

from numpy.random import RandomState
from scipy import linalg
from scipy import sparse
from scipy.sparse.linalg import eigs

__all__ = [
    "fast_spectral_initialization",
    "generate_internal_weights",
    "generate_input_weights"
]


def _is_probability(proba):
    return 1. - proba >= 0. and proba >= 0.


def _get_random_state(seed):
    if isinstance(seed, RandomState):
        return seed
    else:
        return RandomState(seed)


def fast_spectral_initialization(N: int,
                                 spectral_radius: float = None,
                                 proba: float = 0.1,
                                 seed: Union[int, RandomState] = None,
                                 verbose: bool = False,
                                 sparsity_type: str = 'csr',
                                 typefloat=np.float64):
    """The fast spectral radius (FSI) approach for weights
    initialization introduced in [1].

    This method is well suited for computation and rescaling of
    very large weights matrices, with a number of neurons typically
    above 500-1000.

    Parameters
    ----------
    N : int
        Number of reservoir units, i.e. dimension of
        the square weights matrix.
    spectral_radius : float, optional
        Maximum desired eigenvalue of the
        reservoir weights matrix, by default None.
    proba : float, optional
        Probability of non zero connection,
        density of the weight matrix, by default 0.1
    seed : int or RandomState, optional
        Random state generator seed, for reproducibility,
        by default None
    verbose : bool, optional
    sparsity_type : {"csr", "csc", "coo"} optional
        Scipy sparse matrix format. "csr" by default.
    typefloat : np.dtype, optional

    Returns
    -------
    np.ndarray or scipy.sparse matrix
        A reservoir weights matrix.

    Raises
    ------
    ValueError
        Invalid non zero connection probability.

    References
    -----------

        [1] C. Gallicchio, A. Micheli, L. Pedrelli.
        Fast Spectral Radius Initialization for Recurrent
        Neural Networks. (2020)

    Example
    -------

        >>> from reservoirpy.mat_gen import fast_spectral_initialization
        >>> W = fast_spectral_initialization(5, proba=0.5, seed=42)
        >>> W.toarray()
        array([[ 0.13610996, -0.57035192,  0.        ,  0.        ,  0.        ],
               [ 0.26727631,  0.        , -0.22370579, -0.04951286,  0.        ],
               [ 0.60065244,  0.        , -0.02846888,  0.        ,  0.08786003],
               [ 0.        ,  0.39151551,  0.4157107 ,  0.        ,  0.41754172],
               [ 0.        ,  0.72101228,  0.        ,  0.        ,  0.        ]])
    """
    if not _is_probability(proba):
        raise ValueError(f"proba = {proba} not in [0; 1].")

    rs = _get_random_state(seed)

    if spectral_radius is None or proba == 0.:
        a = 1
    else:
        a = -(6 * spectral_radius) / (np.sqrt(12) * np.sqrt((proba * N)))

    if proba < 1:
        return sparse.random(N, N, density=proba,
                             random_state=rs, format=sparsity_type,
                             data_rvs=lambda s: rs.uniform(a, -a, size=s))

    else:
        return np.random.uniform(a, -a, size=(N, N))


def generate_internal_weights(N: int,
                              spectral_radius: float = None,
                              proba: float = 0.1,
                              Wstd: float = 1.0,
                              sparsity_type: str = 'csr',
                              seed: Union[int, RandomState] = None,
                              verbose: bool = False,
                              typefloat=np.float64):
    """Method that generate the weight matrix that will be used
    for the internal connections of the reservoir.

    Weights will follow a normal distribution of mean 0 and
    scale `Wstd` (by default 1), and can then be rescale to
    reach a specific spectral radius.

    Parameters
    ----------
    N : int
        Number of reservoir units, i.e. dimension of
        the square weights matrix.
    spectral_radius : float, optional
        Maximum desired eigenvalue of the
        reservoir weights matrix, by default None
    proba : float, optional
        Probability of non zero connection,
        density of the weight matrix, by default 0.1
    Wstd : float, optional
        Standard deviation of internal weights, by default 1.0
    sparsity_type : str, optional
        If set ot one of the scipy.sparse matrix format,
        will generate internal weights matrix in a sparse
        format, by default None
    seed : int or RandomState, optional
        Random state generator seed, for reproducibility,
        by default None
    verbose : bool, optional
    typefloat : numpy.dtype, optional

    Returns
    -------
    np.ndarray or scipy.sparse matrix
        A reservoir weights matrix.

    Raises
    ------
    ValueError
        Invalid non zero connection probability.

    Example
    -------

        >>> from reservoirpy.mat_gen import fast_spectral_initialization
        >>> W = generate_internal_weights(5, proba=0.5, seed=42)
        >>> W.toarray()
        array([[-1.72491783, -0.2257763 ,  0.        ,  0.        ,  0.        ],
               [-1.4123037 ,  0.        , -1.01283112, -1.91328024,  0.        ],
               [ 0.0675282 ,  0.        , -1.42474819,  0.        ,  1.46564877],
               [ 0.        ,  0.24196227, -0.90802408,  0.        , -0.56228753],
               [ 0.        ,  0.31424733,  0.        ,  0.        ,  0.        ]])
    """
    if not _is_probability(proba):
        raise ValueError(f"proba = {proba} not in [0; 1].")

    rs = _get_random_state(seed)

    if sparsity_type is not None:
        w = sparse.random(N, N, density=proba, format=sparsity_type,
                          random_state=rs,
                          data_rvs=lambda s: rs.normal(0, Wstd, size=s))
        rhoW = max(abs(eigs(w, k=1, which='LM', return_eigenvectors=False)))
    else:
        mask = 1 * (rs.rand(N, N) < proba)
        mat = rs.normal(0, Wstd, (N, N))
        w = np.multiply(mat, mask, dtype=typefloat)

        # Computing the spectral radius of W matrix
        rhoW = max(abs(linalg.eig(w)[0]))

    if verbose:
        print("Spectra radius of generated matrix before "
              "applying another spectral radius: " + str(rhoW))

    if spectral_radius is not None:
        w *= spectral_radius / rhoW
        if verbose:
            if sparse.issparse(w):
                rhoW_after = max(abs(sparse.linalg.eigs(w,
                                                        k=1,
                                                        which='LM',
                                                        return_eigenvectors=False)))
            else:
                rhoW_after = max(abs(linalg.eig(w)[0]))
            print("Spectra radius matrix after applying "
                  "another spectral radius: " + str(rhoW_after))

    return w


def generate_input_weights(N: int,
                           dim_input: int,
                           input_scaling: float = None,
                           proba: float = 0.1,
                           input_bias: bool = False,
                           seed: Union[int, RandomState] = None,
                           verbose: bool = False,
                           typefloat=np.float64):
    """
    Generate input or feedback weights for the reservoir.

    Weights are drawn from a discrete bimodal distribution,
    i.e. are always equal to 1 or -1. Then, they can be rescaled
    to a specific constant using the `input_scaling` parameter.

    Parameters
    ----------
        N: int
            Number of units in the connected reservoir.
        dim_input: int
            Dimension of the inputs connected to the reservoir.
        input_scaling: float, optional
            Constant value used to rescale the weights.
        input_bias: bool, optional
            If True, will add a row to the matrix to take into
            account a constant bias added to the input.
            Mandatory when using a :py:class:`reservoirpy.ESN` with
            `input_bias` set.
        proba: float, optional
            Probability of non-zero connections, density of
            the matrix, by default 0.1.
        seed : int or RandomState, optional
            Random state generator seed, for reproducibility,
            by default None
        verbose : bool, optional
        typefloat : numpy.dtype, optional
    Returns
    -------
    np.ndarray or scipy.sparse matrix
        A reservoir input or feedback weights matrix.

    Raises
    ------
    ValueError
        Invalid non zero connection probability.

    Example
    -------

        >>> from reservoirpy.mat_gen import generate_input_weights
        >>> Win = generate_input_weights(10, 2, input_scaling=2., proba=0.5, seed=42)
        >>> Win
        array([[-2., -0.],
               [ 0.,  0.],
               [ 2.,  2.],
               [ 2., -0.],
               [ 0.,  0.],
               [-2.,  0.],
               [-0.,  2.],
               [-2.,  2.],
               [ 2., -0.],
               [-2., -2.]])

    """
    if not _is_probability(proba):
        raise ValueError(f"proba = {proba} not in [0; 1].")

    rs = _get_random_state(seed)

    if input_bias:
        dim_input += 1
    mask = 1 * (rs.rand(N, dim_input) < proba)
    mat = rs.randint(0, 2, (N, dim_input)) * 2 - 1
    w = np.multiply(mat, mask, dtype=typefloat)
    if input_scaling is not None:
        w = input_scaling * w

    return w



import os
import time
import json
import warnings
from os import path
from functools import partial
from glob import glob

import numpy as np


def _get_conf_from_json(confpath):
    if not(path.isfile(confpath)):
        raise FileNotFoundError(f"Training conf '{confpath}' not found.")
    else:
        config = {}
        with open(confpath, "r") as f:
            config = json.load(f)
        return _parse_config(config)


def _parse_config(config):

    import hyperopt as hopt

    required_args = ["exp", "hp_max_evals", "hp_method", "hp_space"]
    for arg in required_args:
        if config.get(arg) is None:
            raise ValueError(f"No {arg} argument found in config file.")

    if config["hp_method"] not in ["tpe", "random"]:
        raise ValueError(f"Unknow hyperopt algorithm: {config['hp_method']}. Available algs: random, tpe.")
    else:
        if config["hp_method"] == "random":
            config["hp_method"] = partial(hopt.rand.suggest)
        if config["hp_method"] == "tpe":
            config["hp_method"] = partial(hopt.tpe.suggest)

    space = {}
    for arg, specs in config["hp_space"].items():
        space[arg] = _parse_hyperopt_searchspace(arg, specs)

    config["hp_space"] = space

    return config


def _parse_hyperopt_searchspace(arg, specs):

    import hyperopt as hopt

    if specs[0] == "choice":
        return hopt.hp.choice(arg, specs[1:])
    if specs[0] == "randint":
        return hopt.hp.randint(arg, *specs[1:])
    if specs[0] == "uniform":
        return hopt.hp.uniform(arg, *specs[1:])
    if specs[0] == "quniform":
        return hopt.hp.quniform(arg, *specs[1:])
    if specs[0] == "loguniform":
        return hopt.hp.loguniform(arg, np.log(specs[1]), np.log(specs[2]))
    if specs[0] == "qloguniform":
        return hopt.hp.qloguniform(arg, np.log(specs[1]), np.log(specs[2]), specs[3])
    if specs[0] == "normal":
        return hopt.hp.normal(arg, *specs[1:])
    if specs[0] == "qnormal":
        return hopt.hp.qnormal(arg, *specs[1:])
    if specs[0] == "lognormal":
        return hopt.hp.lognormal(arg, np.log(specs[1]), np.log(specs[2]))
    if specs[0] == "qlognormal":
        return hopt.hp.qlognormal(arg, np.log(specs[1]), np.log[specs[2]], specs[3])


def _get_report_path(exp_name, base_path=None):

    base_path = '.' if base_path is None else base_path

    report_path = path.join(base_path, exp_name, 'results')

    if not(path.isdir(base_path)):
        os.mkdir(base_path)

    if not(path.isdir(path.join(base_path, exp_name))):
        os.mkdir(path.join(base_path, exp_name))

    if not(path.isdir(report_path)):
        os.mkdir(report_path)

    return report_path


def research(objective, dataset, config_path, report_path=None):
    """
    Wrapper for hyperopt fmin function. Will run hyperopt fmin on the
    objective function passed as argument, on the data stored in the
    dataset argument.

    Arguments:
    ----------
        objective {Callable} -- Objective function defining the function to
                                optimize. Must be able to receive the dataset argument and
                                all parameters sampled by hyperopt during the search. These
                                parameters must be keyword arguments only without default value
                                (this can be achieve by separating them from the other arguments
                                with an empty starred expression. See examples for more info.)
        dataset {Any} -- Argument used to pass data to the objective function during the hyperopt run.
        config_path {str or Path} -- Path to the hyperopt experimentation configuration file used to
                                     define this run.
    Keywords arguments:
    -------------------
        report_path {str} -- Path to the directory where to store the results of the run. By default,
                             this directory is set to be {name of the experiment}/results/. [default: None]
    """
    import hyperopt as hopt

    config = _get_conf_from_json(config_path)
    report_path = _get_report_path(config["exp"], report_path)

    def objective_wrapper(kwargs):

        try:
            start = time.time()

            returned_dict = objective(dataset, config, **kwargs)

            end = time.time()
            duration = end - start

            returned_dict['status'] = hopt.STATUS_OK
            returned_dict['start_time'] = start
            returned_dict['duration'] = duration

            save_file = f"{returned_dict['loss']:.7f}_hyperopt_results"

        except Exception as e:
            raise e
            start = time.time()

            returned_dict = {
                'status': hopt.STATUS_FAIL,
                'start_time': start,
                'error': str(e),
            }

            save_file = f"ERR{start}_hyperopt_results"

        try:
            json_dict = {'returned_dict': returned_dict, 'current_params': kwargs}
            save_file = path.join(report_path, save_file)
            nb_save_file_with_same_loss = len(glob(f"{save_file}*"))
            save_file = f"{save_file}_{nb_save_file_with_same_loss+1}call.json"
            with open(save_file, "w+") as f:
                json.dump(json_dict, f)
        except Exception as e:
            warnings.warn("Results of current simulation were NOT saved "
                          "correctly to JSON file.")
            warnings.warn(str(e))

        return returned_dict

    search_space = config["hp_space"]

    trials = hopt.Trials()

    if config.get("seed") is None:
        rs = np.random.RandomState()
    else:
        rs = np.random.RandomState(config["seed"])

    best = hopt.fmin(objective_wrapper,
                     space=search_space,
                     algo=config["hp_method"],
                     max_evals=config['hp_max_evals'],
                     trials=trials,
                     rstate=rs)

    return best, trials



import os
import math
import json
import time

from os import path

import numpy as np



def get_results(exp):
    report_path = path.join(exp, "results")
    results = []
    for file in os.listdir(report_path):
        if path.isfile(path.join(report_path, file)):
            with open(path.join(report_path, file), 'r') as f:
                results.append(json.load(f))
    return results


def outliers_idx(values, max_deviation):
    mean = values.mean()

    dist = abs(values - mean)
    corrected = dist < mean + max_deviation
    return corrected


def logscale_plot(ax, xrange, yrange, base=10):

    from matplotlib import ticker

    if xrange is not None:
        ax.xaxis.set_minor_formatter(ticker.LogFormatter())
        ax.xaxis.set_major_formatter(ticker.LogFormatter())
        ax.set_xscale("log", base=base)
        ax.set_xlim([np.min(xrange), np.max(xrange)])
    if yrange is not None:
        ax.yaxis.set_minor_formatter(ticker.LogFormatter())
        ax.yaxis.set_major_formatter(ticker.LogFormatter())
        ax.set_yscale("log", base=base)
        ax.set_ylim([yrange.min() - 0.1*yrange.min(), yrange.max() + 0.1*yrange.min()])


def scale(x):
    return (x - x.min()) / (x.ptp())


def cross_parameter_plot(ax, values, scores, loss, smaxs, cmaxs, lmaxs, p1, p2, log1, log2, cat1, cat2):

    X = values[p2].copy()
    Y = values[p1].copy()

    to_log = []
    if log1 and not cat2:
        to_log.append(X)
    else:
        to_log.append(None)
        if cat2:
            ax.margins(x=0.05)
    if log1 and not cat1:
        to_log.append(Y)
    else:
        to_log.append(None)

    logscale_plot(ax, *to_log)
    # if log2 and not cat2:
    #     logscale_plot(ax, None, Y)

    ax.tick_params(axis='both', which='both')
    ax.tick_params(axis='both', labelsize="xx-small")
    ax.grid(True, which="both", ls="-", alpha=0.75)

    ax.set_xlabel(p2)
    ax.set_ylabel(p1)

    sc_l = ax.scatter(X[lmaxs], Y[lmaxs], scores[lmaxs]*100, c=loss[lmaxs], cmap="inferno")
    sc_s = ax.scatter(X[smaxs], Y[smaxs], scores[smaxs]*100, c=cmaxs, cmap="YlGn")
    sc_m = ax.scatter(X[~(lmaxs)], Y[~(lmaxs)], scores[~(lmaxs)]*100, color="red")

    return sc_l, sc_s, sc_m


def loss_plot(ax, values, scores, loss, smaxs, cmaxs, lmaxs, p, log, categorical, legend, loss_behaviour):

    X = values[p].copy()

    if log and not(categorical):
        logscale_plot(ax, X, loss)
    else:
        logscale_plot(ax, None, loss)
        if categorical:
            ax.margins(x=0.05)

    ax.set_xlabel(p)
    ax.set_ylabel("loss")

    ax.tick_params(axis='both', which='both')
    ax.tick_params(axis='both', labelsize="xx-small")
    ax.grid(True, which="both", ls="-", alpha=0.75)

    sc_l = ax.scatter(X[lmaxs], loss[lmaxs], scores[lmaxs]*100, color="orange")
    sc_s = ax.scatter(X[smaxs], loss[smaxs], scores[smaxs]*100, c=cmaxs, cmap="YlGn")
    if loss_behaviour == "min":
        sc_m = ax.scatter(X[~(lmaxs)], [loss.min()]*np.sum(~lmaxs), scores[~(lmaxs)]*100, color="red", label="Loss min.")
    else:
        sc_m = ax.scatter(X[~(lmaxs)], [loss.max()]*np.sum(~lmaxs), scores[~(lmaxs)]*100, color="red", label="Score max.")

    if legend:
        ax.legend()

    return sc_l, sc_s, sc_m


def parameter_violin(ax, values, scores, loss, smaxs, cmaxs, p, log, legend):

    import matplotlib.pyplot as plt

    y = values[p].copy()[smaxs]
    all_y = values[p].copy()

    if log:
        y = np.log10(y)
        all_y = np.log10(all_y)

    ax.get_yaxis().set_ticks([])
    ax.tick_params(axis='x', which='both')

    ax.xaxis.set_major_locator(plt.MultipleLocator(1))

    def format_func(value, tick_number):
        return "$10^{"+str(int(np.floor(value)))+"}$"

    ax.xaxis.set_major_formatter(plt.FuncFormatter(format_func))

    ax.set_xlabel(p)
    ax.grid(True, which="both", ls="-", alpha=0.75)
    ax.set_xlim([np.min(all_y), np.max(all_y)])

    violin = ax.violinplot(y, vert=False, showmeans=False, showextrema=False)

    for pc in violin['bodies']:
        pc.set_facecolor('forestgreen')
        pc.set_edgecolor('white')

    quartile1, medians, quartile3 = np.percentile(y, [25, 50, 75])

    ax.scatter(medians, 1, marker='o', color='orange', s=30, zorder=4, label="Median")
    ax.hlines(1, quartile1, quartile3, color='gray', linestyle='-', lw=4, label="Q25/Q75")
    ax.vlines(y.mean(), 0.5, 1.5, color='blue', label="Mean")

    ax.scatter(np.log10(values[p][scores.argmax()]), 1, color="red", zorder=5, label="Best score")
    ax.scatter(y, np.ones_like(y), c=cmaxs, cmap="YlGn", alpha=0.5, zorder=3)

    if legend:
        ax.legend(loc=2)


def parameter_bar(ax, values, scores, loss, smaxs, cmaxs, p, categories):

    import matplotlib.pyplot as plt

    y = values[p].copy()[smaxs]

    ax.set_xlabel(p)
    ax.grid(True, which="both", ls="-", alpha=0.75)

    heights = []
    for p in categories:
        heights.append(y.tolist().count(p))

    bar = ax.bar(x=categories, height=heights, color="forestgreen", alpha=0.3)


def plot_hyperopt_report(exp, params, metric='loss', loss_metric="loss",
                         loss_behaviour="min", not_log=None, categorical=None,
                         max_deviation=None, title=None):
    """Cross paramater scatter plot of hyperopt trials.

    Installation of Matplotlib and Seaborn packages is required to use this tool.

    Arguments:
        exp {str or Path} -- Report directory storing hyperopt trials results.
        params {list} -- Parameters to plot.

    Keyword Arguments:
        metric {str} -- Metric to use as performance measure, stored in the hyperopt trials results dictionnaries.
                        May be different from loss metric. By default, loss is used as performance metric. (default: 'loss')
        not_log {list} -- Parameters to plot with a linear scale. By default, all scales are logarithmic.
        categorical {list} -- List of categorical parameters. (default: None)
        max_deviation {float} -- Maximum standard deviation expected from the loss mean. Useful to remove outliers.
                                 (default: {None})
        title {str} -- Optional title for the figure. (default: {None})

    Returns:
        [matplotlib.pyplot.figure] -- Matplotlib figure object.

    """

    import matplotlib.pyplot as plt
    import seaborn as sns

    sns.set(context="paper", style="darkgrid", font_scale=1.5)
    N = len(params)
    not_log = not_log or []

    results = get_results(exp)

    loss = np.array([r['returned_dict'][loss_metric] for r in results])
    scores = np.array([r['returned_dict'][metric] for r in results])

    if max_deviation is not None:
        not_outliers = outliers_idx(loss, max_deviation)
        loss = loss[not_outliers]
        scores = scores[not_outliers]
        values = {p: np.array([r['current_params'][p] for r in results])[not_outliers] for p in params}
    else:
        values = {p: np.array([r['current_params'][p] for r in results]) for p in params}

    categorical = categorical or []

    for p in categorical:
        values[p] = values[p].astype(str)

    # Sorting for categorical plotting
    all_categorical = [val for p, val in values.items() if p in categorical]
    all_numerical = [val for p, val in values.items() if p not in categorical]

    sorted_idx = np.lexsort((loss, scores, *all_numerical, *all_categorical))

    loss = np.array([loss[i] for i in sorted_idx])
    scores = np.array([scores[i] for i in sorted_idx])

    for p, val in values.items():
        values[p] = np.array([val[i] for i in sorted_idx])

    scores = scale(scores)

    ## loss and f1 values

    if loss_behaviour == "min":
        lmaxs = loss > loss.min()
    else:
        lmaxs = loss < loss.max()

    c_l = np.log10(loss[lmaxs])

    percent = math.ceil(len(scores) * 0.05)
    smaxs = scores.argsort()[-percent:][::-1]
    cmaxs = scale(scores[smaxs])

    ## gridspecs

    fig = plt.figure(figsize=(15, 19), constrained_layout=True)
    gs = fig.add_gridspec(2, 1, height_ratios=[2/30, 28/30])
    fig.suptitle(f"Hyperopt trials summary - {title}", size=15)

    gs0 = gs[0].subgridspec(1, 3)
    gs1 = gs[1].subgridspec(N + 1, N)

    lbar_ax = fig.add_subplot(gs0[0, 0])
    fbar_ax = fig.add_subplot(gs0[0, 1])
    rad_ax = fig.add_subplot(gs0[0, 2])
    rad_ax.axis('off')

    # plot
    axes = []
    for i, p1 in enumerate(params):
        for j, p2 in enumerate(params):
            ax = fig.add_subplot(gs1[i, j])
            axes.append(ax)
            if p1 == p2:
                sc_l, sc_s, sc_m = loss_plot(ax, values, scores, loss, smaxs,
                                             cmaxs, lmaxs, p2, not(p2 in not_log),
                                             p2 in categorical, (i==0 and j==0), 
                                             loss_behaviour)
            else:
                sc_l, sc_s, sc_m = cross_parameter_plot(ax, values, scores, loss,
                                     smaxs, cmaxs, lmaxs, p1, p2,
                                     not(p1 in not_log), not(p2 in not_log),
                                     p1 in categorical, p2 in categorical)

    #legends

    handles, labels = sc_l.legend_elements(prop="sizes")
    legend = rad_ax.legend(handles, labels, loc="center left",
                           title=f"Normalized {metric} (%)", mode='expand',
                           ncol=len(labels) // 2 + 1)

    l_cbar = fig.colorbar(sc_l, cax=lbar_ax, ax=axes, orientation="horizontal")
    _ = l_cbar.ax.set_title("Loss value")

    f_cbar = fig.colorbar(sc_s, cax=fbar_ax, ax=axes,
                          orientation="horizontal", ticks=[0, 0.5, 1])
    _ = f_cbar.ax.set_title(f"{metric} best population")
    _ = f_cbar.ax.set_xticklabels(["5% best", "2.5% best", "Best"])

    #violinplots

    legend = True
    for i, p in enumerate(params):
        ax = fig.add_subplot(gs1[-1, i])
        if p in categorical:
            parameter_bar(ax, values, scores,
                          loss, smaxs, cmaxs, p, sorted(list(set(values[p]))))
        else:
            parameter_violin(ax, values, scores,
                            loss, smaxs, cmaxs, p, not(p in not_log), legend)
            legend = False
        if legend:
            ax.set_ylabel(f"5% best {metric}\nparameter distribution")
    return fig

if __name__ == "__main__":

    import matplotlib.pyplot as plt

    fig = plot_hyperopt_report("/home/nathan/Bureau/canary/16syll-decoder/hyperopt/16syll-ot-8e-noise-seed-ridge",
                         ["seed", "ridge"], metric='F1', not_log=None, categorical=["seed"], max_deviation=None, title=None,
                         loss_behaviour="max")

    fig.savefig("test.png")


import numpy as np

# TODO: major refactorization: module name, backend handler, doc


def get_function(name):  # pragma: no cover
    index = {
        "softmax": softmax,
        "softplus": softplus,
        "sigmoid": sigmoid,
        "tanh": tanh,
        "identity": identity,
        "relu": relu,
        "smax": softmax,
        "sp": softplus,
        "sig": sigmoid,
        "id": identity,
        "re": relu,
    }

    if index.get(name) is None:
        raise ValueError(f"Function name must be one of {[k for k in index.keys()]}")
    else:
        return index[name]


def elementwise(func):
    vect = np.vectorize(func)

    def vect_wrapper(x):
        u = np.asanyarray(x)
        return vect(u)

    return vect_wrapper


def softmax(x):
    return np.exp(x) / np.exp(x).sum()


@elementwise
def softplus(x):
    return np.log(1 + np.exp(x))


@elementwise
def sigmoid(x):
    if x < 0:
        u = np.exp(x)
        return u / (u + 1)
    return 1 / (1 + np.exp(-x))


def tanh(x):
    return np.tanh(x)


@elementwise
def identity(x):
    return x


@elementwise
def relu(x):
    if x < 0:
        return 0
    return x



from typing import Tuple

import numpy as np

from scipy.integrate import solve_ivp



def rabinovich_fabrikant(n_timesteps: int, state0: list=[-1, 0, 0.5], 
                         gamma: float=0.89, alpha: float=1.1, h: float=0.05)-> Tuple[np.ndarray, np.ndarray]:
    """Rabinovitch-Fabrikant equations timeseries.
    .. math::
        \dv{x}{t} = y(z - 1 + x^2) + \gamma x
    .. math::
        \dv{y}{t} = x(3z + 1 - x^2) + \gamma y
    .. math::
        \dv{z}{t} = -2z(\alpha + xy)
    Args:
        n_timesteps (int): Number of timesteps to generate.
        state0 (list, optional): Initial conditions of the system. Defaults to [-1, 0, 0.5].
        gamma (float, optional): Gamma parameter of the equations. Defaults to 0.89.
        alpha (float, optional): Alpha parameter of the equations. Defaults to 1.1.
        h (float, optional): Control the time delta between two timesteps. Defaults to 0.05.
    Returns:
        Tuple[np.ndarray, np.ndarray]: Generated timeseries and timesteps.
    """
    
    def rabinovich_fabrikant_diff(t, state):
        x, y, z = state
        dx = y*(z - 1 + x**2) + gamma*x
        dy = x*(3*z + 1 - x**2) + gamma*y
        dz = -2*z*(alpha + x*y)
        return dx, dy, dz

    t = np.arange(0.0, n_timesteps*h, h)

    sol = solve_ivp(rabinovich_fabrikant_diff, 
                     y0=state0, 
                     t_span=(0.0, n_timesteps*h),
                     dense_output=True)
    
    return sol.sol(t).T, t


# -*- coding: utf-8 -*-
#!/usr/bin/env python -W ignore::DeprecationWarning
"""reservoirpy/esn_online

Echo State Networks with online learning (FORCE learning)
"""
import os
import time
import warnings
import pickle
from typing import Sequence, Callable, Tuple, Union, Dict

import numpy as np
from scipy import linalg
from tqdm import tqdm

from .utils import check_values, _save




class ESNOnline(object):

    def __init__(self,
                 lr: float,
                 W: np.ndarray,
                 Win: np.ndarray,
                 Wout: np.ndarray,
                 alpha_coef: float=1e-6,
                 use_raw_input: bool=False,
                 input_bias: bool=True,
                 Wfb: np.ndarray=None,
                 fbfunc: Callable=None,
                 typefloat: np.dtype=np.float64):
        """Base class of Echo State Networks

        Arguments:
            lr {float} -- Leaking rate
            W {np.ndarray} -- Reservoir weights matrix
            Win {np.ndarray} -- Input weights matrix
            Wout {np.ndarray} -- Output weights matrix

        Keyword Arguments:
            alpha_coef {float} -- Coefficient to scale the inversed state correlation matrix
                                  used for FORCE learning
            use_raw_input {bool} -- If True, input is used directly when computing output
                                    (default: {False})
            input_bias {bool} -- If True, will add a constant bias
                                 to the input vector (default: {True})
            Wfb {np.ndarray} -- Feedback weights matrix
            fbfunc {Callable} -- Feedback activation function. (default: {None})
            typefloat {np.dtype} -- Float precision to use (default: {np.float32})

        Raises:
            ValueError: If a feedback matrix is passed without activation function.
            NotImplementedError: If trying to set input_bias to False. This is not
                                 implemented yet.
        """

        self.W = W
        self.Win = Win
        self.Wfb = Wfb
        self.Wout = Wout

        self.use_raw_inp = use_raw_input

        # check if dimensions of matrices are coherent
        self._autocheck_dimensions()
        self._autocheck_nan()

        self.N = self.W.shape[1] # number of neurons

        self.in_bias = input_bias
        self.dim_inp = self.Win.shape[1] - 1 if self.in_bias else self.Win.shape[1] # dimension of inputs (not including the bias at 1)

        self.dim_out = Wout.shape[0]
        self.state_size = Wout.shape[1]
        self.output_values = np.zeros((self.dim_out,1)).astype(typefloat)

        self.typefloat = typefloat
        self.lr = lr # leaking rate

        self.fbfunc = fbfunc
        if self.Wfb is not None and self.fbfunc is None:
            raise ValueError(f"If a feedback matrix is provided, \
                fbfunc must be a callable object, not {self.fbfunc}.")

        self.alpha_coef = alpha_coef # coef used to init state_corr_inv matrix

        # Init internal state vector and state_corr_inv matrix
        # (useful if we want to freeze the online learning)
        self.reset_reservoir()
        self.reset_correlation_matrix()


    def __repr__(self):
        trained = True
        if self.Wout is None:
            trained = False
        fb = True
        if self.Wfb is None:
            fb=False
        out = f"ESN(trained={trained}, feedback={fb}, N={self.N}, "
        out += f"lr={self.lr}, input_bias={self.in_bias}, input_dim={self.N})"
        return out


    def _autocheck_nan(self):
        """ Auto-check to see if some important variables do not have a problem (e.g. NAN values). """
        # assert np.isnan(self.W).any() == False, "W matrix should not contain NaN values."
        assert np.isnan(self.Win).any() == False, "Win matrix should not contain NaN values."
        if self.Wfb is not None:
            assert np.isnan(self.Wfb).any() == False, "Wfb matrix should not contain NaN values."


    def _autocheck_dimensions(self):
        """ Auto-check to see if ESN matrices have correct dimensions."""
        # W dimensions check list
        assert len(self.W.shape) == 2, f"W shape should be (N, N) but is {self.W.shape}."
        assert self.W.shape[0] == self.W.shape[1], f"W shape should be (N, N) but is {self.W.shape}."

        # Win dimensions check list
        assert len(self.Win.shape) == 2, f"Win shape should be (N, input) but is {self.Win.shape}."
        err = f"Win shape should be ({self.W.shape[1]}, input) but is {self.Win.shape}."
        assert self.Win.shape[0] == self.W.shape[0], err

        # Wout dimensions check list
        assert len(self.Wout.shape) == 2, f"Wout shape should be (output, nb_states) but is {self.Wout.shape}."
        nb_states = self.Win.shape[1] + self.W.shape[0] + 1 if self.use_raw_inp else self.W.shape[0] + 1
        err = f"Wout shape should be (output, {nb_states}) but is {self.Wout.shape}."
        assert self.Wout.shape[1] == nb_states, err

        # Wfb dimensions check list
        if self.Wfb is not None:
            assert len(self.Wfb.shape) == 2, f"Wfb shape should be (input, output) but is {self.Wfb.shape}."
            err = f"Wfb shape should be ({self.Win.shape[0]}, {self.Wout.shape[0]}) but is {self.Wfb.shape}."
            assert (self.Win.shape[0],self.Wout.shape[0]) == self.Wfb.shape, err


    def _autocheck_io(self,
                      inputs,
                      outputs=None):

        # Check if inputs and outputs are lists
        assert type(inputs) is list, "Inputs should be a list of numpy arrays"
        if outputs is not None:
            assert type(outputs) is list, "Outputs should be a list of numpy arrays"

        # check if Win matrix has coherent dimensions with input dimensions
        inputs_0 = inputs[0]
        if self.in_bias:
            err = f"With bias, Win matrix should be of shape ({self.N}, "
            err += f"{inputs_0.shape[0] + 1}) but is {self.Win.shape}."
            assert self.Win.shape[1] == inputs_0.shape[0] + 1, err
        else:
            err = f"Win matrix should be of shape ({self.N}, "
            err += f"{inputs_0.shape[0]}) but is {self.Win.shape}."
            assert self.Win.shape[1] == inputs_0.shape[0], err

        if outputs is not None:
            # check feedback matrix
            if self.Wfb is not None:
                outputs_0 = outputs[0]
                err = f"With feedback, Wfb matrix should be of shape ({self.N}, "
                err += f"{outputs_0.shape[0]}) but is {self.Wfb.shape}."
                assert outputs_0.shape[0] == self.Wfb.shape[1], err
            # check output weights matrix
            if self.Wout is not None:
                outputs_0 = outputs[0]
                err = f"Wout matrix should be of shape ({outputs_0.shape[0]}, "
                err += f"{self.state_size}) but is {self.Wout.shape}."
                assert (outputs_0.shape[0], self.state_size) == self.Wout.shape, err


    def _get_next_state(self,
                        single_input: np.ndarray) -> np.ndarray:
        """Given a state vector x(t) and an input vector u(t), compute the state vector x(t+1).

        Arguments:
            single_input {np.ndarray} -- Input vector u(t).

        Raises:
            RuntimeError: feedback is enabled but no feedback vector is available.

        Returns:
            np.ndarray -- Next state s(t+1).
        """

        # check if feedback weights matrix is not None but empty feedback
        if self.Wfb is not None and self.output_values is None:
            raise RuntimeError("Missing a feedback vector.")

        x = self.state[1:self.N+1]

        # add bias
        if self.in_bias:
            u = np.hstack((1, single_input)).astype(self.typefloat)
        else:
            u = single_input

        # linear transformation
        x1 = self.Win @ u.reshape(-1, 1) + self.W @ x

        # add feedback if requested
        if self.Wfb is not None:
            x1 += self.Wfb @ self.fbfunc(self.output_values)

        # previous states memory leak and non-linear transformation
        x1 = (1-self.lr)*x + self.lr*np.tanh(x1)

        # return the next state computed
        if self.use_raw_inp:
            self.state = np.vstack((1.0, x1, u.reshape(-1, 1)))
        else:
            self.state = np.vstack((1.0, x1))

        return self.state.copy()


    def compute_output_from_current_state(self):
        """ Compute output from current state s(t) of the reservoir

        Returns:
            np.ndarray -- Output at time t
        """

        assert self.Wout is not None, "Matrix Wout is not initialized/trained yet"

        self.output_values = (self.Wout @ self.state).astype(self.typefloat)
        return self.output_values.copy().ravel()


    def compute_output(self,
                       single_input: np.ndarray,
                       wash_nr_time_step: int=0):
        """ Compute output from input to the reservoir

        Arguments:
            single_input {np.ndarray} -- Input vector u(t).

        Keyword Arguments:
            wash_nr_time_step {int} -- Time for reservoir to run in free (without collecting output
                                       or fitting Wout). (default: {0})

        Returns:
            state {np.ndarray} -- New state after input u(t) is passed
            output {np.ndarray} -- Output after input u(t) is passed
        """

        state = self._get_next_state(single_input)
        output = self.compute_output_from_current_state()

        return output, state


    def reset_reservoir(self):
        """
            Reset reservoir by setting internal values to zero
        """
        self.state = np.zeros((self.state_size,1),dtype=self.typefloat)


    def reset_correlation_matrix(self):
        self.state_corr_inv = np.asmatrix(np.eye(self.state_size)) / self.alpha_coef


    def train_from_current_state(self, targeted_output, indexes = None):
        """ Train Wout from current internal state.

        Arguments:
            error {np.ndarray} -- Error when using current Wout to compute output

        Keyword Arguments:
            indexes {int | list} -- If indexes is not None, only the provided output
                                    indexes are learned (default: {None})
        """

        error = self.output_values - targeted_output.reshape(-1,1)

        self.state_corr_inv = _new_correlation_matrix_inverse(self.state, self.state_corr_inv)

        if indexes == None:
            self.Wout -= error @ (self.state_corr_inv @ self.state).T
        else:
            self.Wout[indexes] -= error[indexes] * (self.state_corr_inv @ self.state).T


    def train(self,
              inputs: Sequence[np.ndarray],
              teachers: Sequence[np.ndarray],
              wash_nr_time_step: int=0,
              verbose: bool=False) -> Sequence[np.ndarray]:
        """Train the ESN model on a sequence of inputs.

        Arguments:
            inputs {Sequence[np.ndarray]} -- Training set of inputs.
            teachers {Sequence[np.ndarray]} -- Training set of ground truth.

        Keyword Arguments:
            wash_nr_time_step {int} -- Number of states to considered as transitory
                            when training. (default: {0})
            verbose {bool} -- if `True`, display progress in stdout.

        Returns:
            Sequence[np.ndarray] -- All states computed, for all inputs.
        """
        inputs_concat = [inp[t,:] for inp in inputs for t in range(inp.shape[0])]
        teachers_concat = [tea[t,:] for tea in teachers for t in range(tea.shape[0])]

        ## Autochecks of inputs and outputs
        self._autocheck_io(inputs=inputs_concat, outputs=teachers_concat)

        if verbose:
            steps = np.sum([i.shape[0] for i in inputs])
            print(f"Training on {len(inputs)} inputs ({steps} steps)-- wash: {wash_nr_time_step} steps")

        # List of all internal states when training
        all_states = []
        start = 1 if self.in_bias else 0
        end = self.N + start

        for i in range(len(inputs)):

            t = 0
            all_states_inp_i = []

            # First 'warm up' the network
            while t < wash_nr_time_step:
                self.compute_output(inputs_concat[i+t])
                t += 1

            # Train Wout on each input
            while t < inputs[i].shape[0]:
                _, state = self.compute_output(inputs_concat[i+t])
                self.train_from_current_state(teachers_concat[i+t])
                all_states_inp_i.append(state[start:end])
                t += 1

            # Pack in all_states
            all_states.append(np.hstack(all_states_inp_i))

        # return all internal states
        return [st.T for st in all_states]


    def run(self,
            inputs: Sequence[np.ndarray],
            verbose: bool=False) -> Tuple[Sequence[np.ndarray], Sequence[np.ndarray]]:
        """Run the model on a sequence of inputs, and returned the states and
           readouts vectors.

        Arguments:
            inputs {Sequence[np.ndarray]} -- Sequence of inputs.

        Keyword Arguments:
            verbose {bool} -- if `True`, display progress in stdout.

        Returns:
            Tuple[Sequence[np.ndarray], Sequence[np.ndarray]] -- All states and readouts,
                                                                 for all inputs.

        """

        inputs_concat = [inp[t,:] for inp in inputs for t in range(inp.shape[0])]

        steps = np.sum([i.shape[0] for i in inputs])
        if verbose:
            print(f"Running on {len(inputs)} inputs ({steps} steps)")

        ## Autochecks of inputs
        self._autocheck_io(inputs=inputs_concat)

        all_outputs = []
        all_states = []
        for i in range(len(inputs)):
            internal_pred = []; output_pred = []
            for t in range(inputs[i].shape[0]):
                output, state = self.compute_output(inputs_concat[i+t])
                internal_pred.append(state)
                output_pred.append(output)
            all_states.append(np.asarray(internal_pred))
            all_outputs.append(np.asarray(output_pred))

        # return all_outputs, all_int_states
        return all_outputs, all_states


    def save(self, directory: str):
        """Save the ESN to disk.

        Arguments:
            directory {str or Path} -- Directory of the saved model.
        """
        _save(self, directory)


    def describe(self) -> Dict:
        """
        Provide descriptive stats about ESN matrices.

        Returns:
            Dict -- Descriptive data.
        """

        desc = {
            "Win": {
                "max": np.max(self.Win),
                "min": np.min(self.Win),
                "mean": np.mean(self.Win),
                "median": np.median(self.Win),
                "std": np.std(self.Win)
            },
            "W": {
                "max": np.max(self.W),
                "min": np.min(self.W),
                "mean": np.mean(self.W),
                "median": np.median(self.W),
                "std": np.std(self.W),
                "sr": max(abs(linalg.eig(self.W)[0]))
            }
        }
        if self.Wfb is not None:
            desc["Wfb"] = {
                "max": np.max(self.Wfb),
                "min": np.min(self.Wfb),
                "mean": np.mean(self.Wfb),
                "median": np.median(self.Wfb),
                "std": np.std(self.Wfb)
            }
        if self.Wout is not None:
            desc["Wout"] = {
                "max": np.max(self.Wout),
                "min": np.min(self.Wout),
                "mean": np.mean(self.Wout),
                "median": np.median(self.Wout),
                "std": np.std(self.Wout)
            }
        return desc




def _new_correlation_matrix_inverse(new_data, old_corr_mat_inv):
    """
        If old_corr_mat_inv is an approximation for the correlation
        matrix inverse of a dataset (p1, ..., pn), then the function
        returns an approximatrion for the correlation matrix inverse
        of dataset (p1, ..., pn, new_data)

        TODO : add forgetting parameter lbda
    """

    P = old_corr_mat_inv
    x = new_data

    # TODO : numerical instabilities if xTP is not computed first (order of multiplications)
    xTP = x.T @ P
    P = P - (P @ x @ xTP)/(1. + np.dot(xTP, x))

    return P



"""ReservoirPy chaotic timeseries generator.

All timeseries defined by differential equations on a continuous space are approximated
using 4-5th order Runge-Kuta method, either homemade (for Mackey-Glass timeseries) 
or from Scipy solve_ivp tool.

Available chaotic attractors:

    Discrete timeseries:
        logistic_map: Logistic map timeseries.
        henon_map: Hénon map timeseries.
        
    Approximations of continuous timeseries:
        mackey_glass: Mackey-Glass delayed differential equations timeseries.
        lorenz: Lorenz system timeseries.
        multiscroll: Double scroll attractor timeseries.
        rabinovitch_fabrikant: Rabinovitch-Fabrikant differential equations timeseries.
"""
from ._mackey_glass import mackey_glass
from ._logistic_map import logistic_map
from ._lorenz import lorenz
from ._multiscroll import multiscroll
from ._rabinovitch_fabrikant import rabinovich_fabrikant
from ._henon_map import henon_map


import numpy as np

from scipy import linalg


def sklearn_linear_model(model):

    def linear_model_solving(X, Y):
            """
                Uses regression method provided during network instanciation to return W such as W * X ~= Ytarget
                First row of X MUST be only ones.
            """
            # Learning of the model (first row of X, which contains only ones, is removed)
            model.fit(X[1:, :].T, Y.T)

            # linear_model provides Matrix A and Vector b such as A * X[1:, :] + b ~= Ytarget
            A = np.asmatrix(model.coef_)
            b = np.asmatrix(model.intercept_).T

            # Then the matrix W = "[b | A]" statisfies "W * X ~= Ytarget"
            return np.asarray(np.hstack([b, A])) 
    
    return linear_model_solving


def ridge_linear_model(ridge, typefloat=np.float32):
    
    def ridge_model_solving(X, Y): 
        ridgeid = (ridge*np.eye(X.shape[0])).astype(typefloat)
        
        return np.dot(np.dot(Y, X.T), linalg.inv(np.dot(X, X.T) + ridgeid))
    
    return ridge_model_solving


def pseudo_inverse_linear_model():
    
    def pseudo_inverse_model_solving(X, Y):
        return np.dot(Y, linalg.pinv(X))
    
    return pseudo_inverse_model_solving


# -*- coding: utf-8 -*-
#!/usr/bin/env python -W ignore::DeprecationWarning
"""
@author: xavier.hinaut #/at\# inria.fr
Copyright Xavier Hinaut 2019-2020
MIT License
"""
import numpy as np
import matplotlib.pyplot as plt

import pickle

def load(path):
    with open(path, 'rb') as f:
        return pickle.load(f)

def byteify(input):
    """
    There's no built-in option to make the json module functions return byte
    strings instead of unicode strings. However, this short and simple recursive
    function will convert any decoded JSON object from using unicode strings to
    UTF-8-encoded byte strings
    Just call this on the output you get from a json.load or json.loads call.

    http://stackoverflow.com/questions/956867/how-to-get-string-objects-instead-of-unicode-ones-from-json-in-python#13105359
    """
    if isinstance(input, dict):
        return {byteify(key):byteify(value) for key,value in input.iteritems()}
    elif isinstance(input, list):
        return [byteify(element) for element in input]
    elif isinstance(input, unicode):
        return input.encode('utf-8')
    else:
        return input

if __name__=='__main__':
    """
    Example how to run it (on 2 lines):
    > python hyperopt_analyse_error.py -l -t your_folder_with_hp_exploration_files/hyperopt_trials.pkl
      -o your_output_folder/error_vs_
    > python hyperopt_analyse_error.py -t hyper-explorations/2019-04-24_19h32__user1_1556127146038513/hyperopt_trials.pkl -o hyper-explorations/2019-04-24_19h32__user1_1556127146038513/error_vs_
    """
    import json
    import argparse

    parser = argparse.ArgumentParser()
    parser.add_argument("-t", "--trials", type=str, help= "file containing all the trials of an experiment")
    parser.add_argument("--title", type = str, help= "title of the figure", default = None)
    parser.add_argument("-o", "--output", type=str, help="file name to save the figure", default = None)
    parser.add_argument("-l", "--logparam", action = "store_true", help= "if set then the axis of the param will be in log scale")
    parser.add_argument("-p", "--plot", action = "store_true", help= "if set then we will plot the images in windows instead of saving them")
    parser.add_argument("-v", "--verbose", action = "store_true", help= "show more information")
    parser.add_argument("-s", "--subset", type=int, default = None, help= "Do not use all trials, but keep the S number of trials indicated by -s S option. This option is useful when you analyse trials saved before the hp search ended.")
    # parser.add_argument("-f", "--filename", type=str, help="json file containing parameters for experiment (root directory is assumed to be params/)")
    # parser.add_argument("-s", "--simulationname", type=str, help="simulation_name: this will be use for the file and saving directory)")
    # parser.add_argument("-l", "--lookup", action='store_true', help="lookup_search_space: Plot the search space given instead of launching a parameter search.")
    args = parser.parse_args()

    print("args.trials", args.trials)
    trials = load(args.trials)
    nr_of_evals = len(trials.trials)
    print("nr of evals:", nr_of_evals)
    if args.subset is not None:
        loss = np.empty((args.subset,))
    else:
        loss = np.empty((nr_of_evals,))
    param = np.empty_like(loss)

    # print("par", trials.trials["misc"]["vals"]

    ss_params = trials.trials[0]["misc"]["vals"].keys()
    print("params from search space:", ss_params)

    if args.verbose:
        print("keys of trials.result dictionary:", trials.trials[0]["result"].keys())

    # params to plot without log_scale
    params_not_log = ['fb_func','fb_proba','thres_repl_unfrq_w', 'wash_nr_time_step']

    for current_param in ss_params:
        for i, t in enumerate(trials.trials):
            if args.verbose:
                print("i=", i)
            if args.subset is not None and i >= args.subset:
                break
            # if i >= nr_of_evals - 1:
            #     break
            loss[i] = t["result"]["loss"]
            param[i] = t["misc"]["vals"][current_param][0]

        f = plt.figure(figsize = (10,10))
        p = f.add_subplot(1,1,1)
        if args.title != None:
            p.set_title(args.title)
        else:
            p.set_title("error vs. "+current_param)
        p.set_yscale("log")
        if args.logparam:
            if current_param not in params_not_log:
                p.set_xscale("log")
        p.set_xlabel("Param: "+current_param)
        p.set_ylabel("Error")
        if current_param == 'fb_func':
            p.set_xlabel("Param: "+current_param+"\n[heavyside01, linear01, linear]")

        # p.scatter(x = param, y = loss, color = "blue")
        p.scatter(x = param, y = loss)
        if args.plot:
            plt.show()
        else:
            if args.output is not None:
                save_name = args.output+'-'+current_param
            else:
                dir_path_of_pickle_file = '/'.join(args.trials.split('/')[:-1])
                if dir_path_of_pickle_file == '':
                    dir_path_of_pickle_file = '.'
                dir_path_of_pickle_file += '/'
                print("saving directory:", dir_path_of_pickle_file)
                if args.subset:
                    sub_name = str(args.subset)
                else:
                    sub_name = ''
                print("saving under file name:", dir_path_of_pickle_file+sub_name+'trials_error-vs-'+current_param)
                save_name = dir_path_of_pickle_file+sub_name+'trials_error-vs-'+current_param
            f.savefig(save_name)



import os
import time
import json
import warnings
from typing import Sequence, Union, Any

import dill
import numpy as np

from scipy import sparse

import reservoirpy


def check_values(array_or_list: Union[Sequence, np.ndarray], value: Any):
    """ Check if the given array or list contains the given value. """
    if value == np.nan:
        assert np.isnan(array_or_list).any() == False, f"{array_or_list} should not contain NaN values."
    if value == None:
        if type(array_or_list) is list:
            assert np.count_nonzero(array_or_list == None) == 0, f"{array_or_list} should not contain None values."
        elif type(array_or_list) is np.array:
            # None is transformed to np.nan when it is in an array
            assert np.isnan(array_or_list).any() == False, f"{array_or_list} should not contain NaN values."
      
      
def _save(esn, directory: str):
    """ Base utilitary for saving an ESN model, based on the ESN class.
    
    Arguments:
        esn {ESN} -- ESN model to save.
        directory {str or Path} -- Directory to store the model.
    """
    # create new directory
    savedir = os.path.join(directory)
    if not os.path.isdir(savedir):
        os.mkdir(savedir)
    else:
        raise OSError(f"Directory '{savedir}' already exists.")
    
    current_time = time.time()    
    
    # store matrices
    Win_path = f"esn-Win-{current_time}.npy"
    Wout_path = None
    Wfb_path = None
    
    if sparse.issparse(esn.W):
        W_path = f"esn-W-{current_time}.npz"
        sparse.save_npz(os.path.join(savedir, W_path), esn.W)
    else:
        W_path = f"esn-W-{current_time}.npy"
        np.save(os.path.join(savedir, W_path), esn.W)
    np.save(os.path.join(savedir, Win_path), esn.Win)
    
    dim_out = None
    if esn.Wout is not None:
        Wout_path = f"esn-Wout-{current_time}.npy"
        np.save(os.path.join(savedir, Wout_path), esn.Wout)
        dim_out = esn.dim_out
        
    fbfunc = None
    fbfunc_info = None
    if esn.Wfb is not None:
        Wfb_path = f"esn-Wfb-{current_time}.npy"
        np.save(os.path.join(savedir, Wfb_path), esn.Wfb)
        # fbfunc is serialized and stored
        fbfunc_info = {
            "cls": str(esn.fbfunc.__class__),
            "name": esn.fbfunc.__name__
        }
        fbfunc = f"fbfunc_save-{current_time}"
        dim_out = esn.dim_out
        with open(os.path.join(savedir, fbfunc), "wb+") as f:
            dill.dump(esn.fbfunc, f)

    reg_model = {"type": "pinv"}
    if esn.ridge is not None:
        reg_model = {"type": "ridge", "coef": esn.ridge}
    elif esn.sklearn_model is not None:
        reg_model = {"type": "sklearn", "path": f"sklearn_func_save-{current_time}"}
        # reg_model is serialized and stored
        with open(os.path.join(savedir, reg_model), "wb+") as f:
            dill.dump(esn.sklearn_model, f)
    
    # a copy of the ESN class is also serialized.
    # allow to load an ESN without necesseraly using the same version of Reservoirpy.
    cls_path = f"cls_bin-{current_time}"
    with open(os.path.join(savedir, cls_path), "wb+") as f:
        dill.dump(esn.__class__, f)
        
    attr = {
        "cls": esn.__class__.__name__,
        "cls_bin": cls_path,
        "version": reservoirpy.__version__,
        "serial": current_time,
        "attr": {
            "W": W_path,
            "Win": Win_path,
            "Wfb": Wfb_path,
            "Wout": Wout_path,
            "N": esn.N,
            "lr": esn.lr,
            "in_bias": esn.in_bias,
            "dim_inp": esn.dim_inp,
            "dim_out": dim_out,
            "typefloat": esn.typefloat.__name__,
            "reg_model": reg_model,
            "fbfunc": fbfunc
        },
        "misc": {
            "fbfunc_info": fbfunc_info
        }
    }
    
    # save a summary file
    with open(os.path.join(savedir, "esn.json"), "w+") as f:
        json.dump(attr, f)
            
            
def load(directory: str):
    """Load an ESN model.
    
    Arguments:
        directory {str or Path} -- Directrory of the saved model.
    
    Returns:
        [ESN] -- Loaded ESN.
    """
    with open(os.path.join(directory, "esn.json"), "r") as f:
        attr = json.load(f)

    model_attr = attr["attr"]
    
    if os.path.splitext(model_attr["W"])[1] == "npy":
        model_attr["W"] = np.load(os.path.join(directory, model_attr["W"]))
    else:
        model_attr["W"] = sparse.load_npz(os.path.join(directory, model_attr["W"]))
        
    model_attr["Win"] = np.load(os.path.join(directory, model_attr["Win"]))
    
    if model_attr["Wout"] is not None:
        model_attr["Wout"] = np.load(os.path.join(directory, model_attr["Wout"]))
    if model_attr["Wfb"] is not None:
        model_attr["Wfb"] = np.load(os.path.join(directory, model_attr["Wfb"]))
        with open(os.path.join(directory, model_attr["fbfunc"]), "rb") as f:
            model_attr["fbfunc"] = dill.load(f)
    

    if model_attr["reg_model"]["type"] == "ridge":
        model_attr["ridge"] = model_attr["reg_model"]["coef"]
        model_attr["reg_model"] = None
    elif model_attr["reg_model"]["type"] == "sklearn":
        with open(os.path.join(directory, model_attr["reg_model"]["path"]), "rb") as f:
            model_attr["reg_model"] = dill.load(f)
        model_attr["ridge"] = None
    else:
        model_attr["reg_model"] = None
        model_attr["ridge"] = None
    
    model_attr["typefloat"] = np.float64
    
    with open(os.path.join(directory, attr["cls_bin"]), 'rb') as f:
        base_cls = dill.load(f)
   
    model = _new_from_save(base_cls, model_attr)
    
    return model

    
def _new_from_save(base_cls, restored_attr):
    
    obj = object.__new__(base_cls)
    for name, attr in restored_attr.items():
        obj.__setattr__(name, attr)
    obj.reg_model = obj._get_regression_model(obj.ridge, obj.reg_model)
    
    return obj


import pytest
import numpy as np

from numpy.testing import assert_almost_equal
from numpy.testing import assert_array_almost_equal
from scipy import linalg
from scipy import sparse

from reservoirpy.mat_gen import fast_spectral_initialization
from reservoirpy.mat_gen import generate_input_weights
from reservoirpy.mat_gen import generate_internal_weights


@pytest.mark.parametrize("N,dim_input,input_bias,expected", [
    (100, 20, False, (100, 20)),
    (100, 20, True, (100, 21)),
    (20, 100, True, (20, 101)),
])
def test_generate_inputs_shape(N, dim_input, input_bias, expected):

    Win = generate_input_weights(N,
                                 dim_input,
                                 input_bias=input_bias)

    assert Win.shape == expected


@pytest.mark.parametrize("N,dim_input,input_bias", [
    (-1, 10, True),
    (100, -5, False),
])
def test_generate_inputs_shape_exception(N, dim_input, input_bias):
    with pytest.raises(ValueError):
        generate_input_weights(N, dim_input, input_bias=input_bias)


@pytest.mark.parametrize("proba,iss", [
    (0.1, 0.1),
    (1.0, 0.5),
    (0.5, 2.0)
])
def test_generate_inputs_features(proba, iss):

    Win = generate_input_weights(100, 20, input_scaling=iss,
                                 proba=proba, seed=1234)
    Win_noiss = generate_input_weights(100, 20, input_scaling=1.0,
                                       proba=proba, seed=1234)

    result_proba = np.sum(Win != 0.0) / Win.size

    assert_almost_equal(result_proba, proba, decimal=1)

    assert_array_almost_equal(Win / iss, Win_noiss, decimal=3)


@pytest.mark.parametrize("proba,iss", [
    (-1, "foo"),
    (5, 1.0)
])
def test_generate_inputs_features_exception(proba, iss):
    with pytest.raises(Exception):
        generate_input_weights(100, 20, input_scaling=iss,
                               proba=proba)


@pytest.mark.parametrize("N,expected", [
    (100, (100, 100)),
    (-1, Exception),
    ("foo", Exception)
])
def test_generate_internal_shape(N, expected):
    if expected is Exception:
        with pytest.raises(expected):
            generate_internal_weights(N)
    else:
        W = generate_internal_weights(N)
        assert W.shape == expected


@pytest.mark.parametrize("sr,proba", [
    (0.5, 0.1),
    (2.0, 1.0),
])
def test_generate_internal_features(sr, proba):

    W = generate_internal_weights(100, spectral_radius=sr,
                                  proba=proba, seed=1234)

    assert_almost_equal(max(abs(linalg.eig(W)[0])), sr, decimal=2)
    assert_almost_equal(np.sum(W != 0.0) / W.size, proba, decimal=1)


@pytest.mark.parametrize("sr,proba", [
    (0.5, 0.1),
    (2.0, 1.0)
])
def test_generate_internal_sparse(sr, proba):

    W = generate_internal_weights(100, spectral_radius=sr,
                                  proba=proba, sparsity_type="csr")

    rho = max(abs(sparse.linalg.eigs(W, k=1, which='LM', return_eigenvectors=False)))
    assert_almost_equal(rho, sr, decimal=2)

    if sparse.issparse(W):
        assert_almost_equal(np.sum(W.toarray() != 0.0) / W.toarray().size,
                            proba, decimal=1)
    else:
        assert_almost_equal(np.sum(W != 0.0) / W.size, proba, decimal=1)


@pytest.mark.parametrize("sr,proba", [
    (1, -0.5),
    (1, 12),
    ("foo", 0.1)
])
def test_generate_internal_features_exception(sr, proba):
    with pytest.raises(Exception):
        generate_internal_weights(100, spectral_radius=sr,
                                  proba=proba)


@pytest.mark.parametrize("N,expected", [
    (100, (100, 100)),
    (-1, Exception),
    ("foo", Exception)
])
def test_fast_spectral_shape(N, expected):
    if expected is Exception:
        with pytest.raises(expected):
            fast_spectral_initialization(N)
    else:
        W = fast_spectral_initialization(N)
        assert W.shape == expected


@pytest.mark.parametrize("sr,proba", [
    (0.5, 0.1),
    (10., 0.5),
    (1., 1.0),
    (1., 0.0)
])
def test_fast_spectral_features(sr, proba):
    W = fast_spectral_initialization(1000, spectral_radius=sr,
                                     proba=proba, seed=1234)

    rho = max(abs(sparse.linalg.eigs(W, k=1, which='LM', return_eigenvectors=False)))
    assert_almost_equal(rho, sr, decimal=0)

    if 1. - proba < 1e-5:
        assert not sparse.issparse(W)
    if sparse.issparse(W):
        assert_almost_equal(np.sum(W.toarray() != 0.0) / W.toarray().size,
                            proba, decimal=1)
    else:
        assert_almost_equal(np.sum(W != 0.0) / W.size, proba, decimal=1)


@pytest.mark.parametrize("sr,proba", [
    (1, -0.5),
    (1, 12),
    ("foo", 0.1)
])
def test_fast_spectral_features_exception(sr, proba):
    with pytest.raises(Exception):
        fast_spectral_initialization(100, spectral_radius=sr,
                                     proba=proba)



import setuptools

with open("README.md", "r") as fh:
    long_description = fh.read()

setuptools.setup(
    name="reservoirpy",
    version="0.2.1",
    author="Xavier Hinaut",
    author_email="xavier.hinaut@inria.fr",
    description="A simple and flexible code for Reservoir Computing architectures like Echo State Networks.",
    long_description=long_description,
    long_description_content_type="text/markdown",
    url="https://github.com/reservoirpy/reservoirpy",
    download_url='https://github.com/reservoirpy/reservoirpy/archive/v0.2.1.tar.gz',
    packages=setuptools.find_packages(),
    classifiers=[
        "Programming Language :: Python :: 3",
        "License :: OSI Approved :: MIT License",
        "Operating System :: OS Independent",
        "Topic :: Scientific/Engineering :: Artificial Intelligence"
    ],
    python_requires='>=3.6',
    install_requires=[
        "tqdm>=4.43.0",
        "joblib>=0.12",
        "dill>=0.3.0"
        'numpy>=1.15.0',
        'scipy>=1.0.0',
        'joblib>=0.12',
    ]
)



import numpy as np 


def henon_map(n_timesteps: int, a: float=1.4, b: float=0.3, state0: list=[0.0, 0.0])->np.ndarray:
    """Hénon map timeseries:
    .. math::
        x(n+1) = 1 - ax(n)^2 + y(n)
        y(n+1) = bx(n)

    Args:
        n_timesteps (int): 
        a (float, optional): a parameter of the Hénon map. Defaults to 1.4 (classic chaotic Hénon map).
        b (float, optional): b parameter of the Hénon map. Defaults to 0.3 (classic chaotic Hénon map).
        state0 (list, optional): Initial condition of the timeseries. Defaults to [0.5, 0.5].
        
    Returns:
        np.ndarray: Hénon map timeseries.
    """
    states = np.zeros((n_timesteps, 2))
    states[0] = np.asarray(state0)

    for i in range(1, n_timesteps):
        states[i][0] = 1 - a*states[i-1][0]**2 + states[i-1][1]
        states[i][1] = b*states[i-1][0]

    return states


# -*- coding: utf-8 -*-
#!/usr/bin/env python -W ignore::DeprecationWarning
"""
@author: xavier.hinaut #/at\# inria.fr
Copyright Xavier Hinaut 2019-2020
MIT License

Some interesting comments on Hyperopt:
http://stackoverflow.com/questions/24673739/hyperopt-set-timeouts-and-modify-space-during-execution
"""
# imports relative to hyperopt
from hyperopt import fmin, tpe, hp, STATUS_OK, STATUS_FAIL, Trials, space_eval, rand
from hyperopt.pyll.stochastic import sample
from functools import partial
# general imports
import numpy as np
import time
import csv
import json
import pprint
import pickle

""" TODO: KEEP AND MODIFY THESE GLOBAL VARIABLES """
### Variables globales modifiables
SAVE_DIR = "hyper-explorations/"
VERBOSE = False
## Variables globales spécifiques à hyperopt (l'optimisateur bayésien)
HP_MAX_EVALS = 20 # TODO: MAXIMUM NUMBER OF TRIALS THAT WILL BE PERFORMED
HP_WARMUP_EVALS = round(HP_MAX_EVALS/3) # NUMBER OF RANDOM TRIALS TO BOOTSTRAP THE ALGORITHM # Nombre de stimulus aléatoires qui vont être exploré avant que l'optimisation bayésienne se mette en marche !
SAVE_TRIAL_OBJ_EVERY = 3 # # TODO: (change if you want less files produced) THE TRIAL OBJECT (containing all trials) IS SAVED EVERY ... TIMES # On sauvegarde le fichier général tous les ... fois
CURR_EVAL = 0 #INIT OF VARIABLE

""" TODO: MAKE YOUR OWN GLOBAL VARIABLES BELOW:
USE MIN AND MAX SCORE IF YOU NEED YOUR METHOD TO BE OPTIMIZED GIVES A SCORE (instead of a loss) AND YOU NEED TO NORMALIZE IT """
# INSERT YOUR OWN PARAMETERS BELOW >>>
SCORE_MIN = 0 #L'utilisateur voit le score s'afficher entre -50 et +50, le score renvoyer est comme le reste des paramètres codé sur int[0;127]
SCORE_MAX = 127
FICHIER_OPTIM2LIVE = '_PythonToLive.txt' #fichier de sortie dans lequel on va écrire les paramètres à tester
FICHIER_LIVE2OPTIM = '_LiveToPython.txt' #fichier d'entrée dans lequel on va lire les résultats pour les paramètres envoyés précédemment
DELIMITATEUR = " "
WAIT_TIME_READ_FILE = 0.5 # in seconds
# <<< INSERT YOUR OWN PARAMETERS HERE


"""
MAIN FUNCTION which exchange data with Ableton Live module
"""
def function_that_returns_the_loss_given_the_parameters(params):
    """
    TODO: MODIFY THIS FUNCTION TO CALL YOUR METHOD, MODEL, NEURAL NETWORK, etc. THAT NEED TO BE OPTIMIZED
    """

    # Here are done the computations, discrimination, model evaluation, ....
    pass
    score = SCORE_MAX ## dummy score

    # Normaliser le score pour en faire une "erreur" (loss) entre 0 et 1
    score_normalise = 1 - (score - SCORE_MIN)/(SCORE_MAX - SCORE_MIN)

    print("SCORE USER + SCORE NORMALISé :", score, score_normalise)

    return score_normalise, score, volume

def objective_dic(params):
    """
    Objective function that takes a dictionary as argument.
    This is the function the optimizer will try to minimize by sampling parameters that are given in input.
    """


    global SAVE_TRIAL_OBJ_EVERY
    global CURR_EVAL
    CURR_EVAL += 1

    ### saving the trial data every SAVE_TRIAL_OBJ_EVERY
    # we save the trail at the beginning because we cannot save the results of
    #   the current trial during the call of the method we are in.
    if (CURR_EVAL-1)%SAVE_TRIAL_OBJ_EVERY == 0:
        try:
            save(trials, SAVE_DIR+"hyperopt_trials_eval"+str(CURR_EVAL-1)+".pkl")
            # save([t for i, t in enumerate(trials.trials) if i < len(trials.trials) ], SAVE_DIR+"hyperopt_trials_eval"+str(CURR_EVAL)+".pkl")
        except:
            print("!!! WARNING: COULD NOT SAVE THE PARTIAL TRIAL OBJECT.")

    print("")
    print("***      HP: params given to objective function:", params)
    print("         current eval : ", CURR_EVAL)
    print("")

    current_params = params

    """ TODO: we do not need seed for this experiment, but maybe you need to do so"""
    # current_params.update({'seed': seed})
    seed = int(time.time()*10**6)

    print("HP: all current params:", current_params)
    print("")

    """ TODO: INIT EXPERIMENT """
    # compexp = experiment.Experiment(current_params)
    # compexp.launch_exp()
    """ no init needed here, but maybe you need it """

    start_time = time.time()
    try:
        """ TODO: LAUNCH EXPERIMENT & RETRIEVE LOSS and some supplementary params """
        score_norm, score_user, volume_user = function_that_returns_the_loss_given_the_parameters(params)
        # std = None
        """ TODO: GET OTHER USEFUL DETAILS ON EACH TRIAL """
        end_time = time.time()
        run_time = end_time-start_time
        # loss_tmp, time_loss = geht_los_einfach(mean, std, run_time) #geht_los(mean, std, run_time)
        loss_tmp, time_loss = score_norm, None
        returned_dic = {'loss': loss_tmp,
                    'status': STATUS_OK,
                    # -- store other results like this
                    # 'loss_variance': std**2,
                    'eval_time': time.time(),
                    'true_loss': score_norm,
                    'start_time': start_time,
                    'end_time': end_time,
                    'run_time': run_time,
                    'time_loss': time_loss,
                    'score_user': score_user,
                    'volume_user': volume_user,
                    # For info
    #                    'true_loss': type float if you pre-compute a test error for a validation error loss, store it here so that Hyperopt plotting routines can find it,
    #                    'true_loss_variance': type float variance in test error estimator,
    #                    'other_stuff': {'list_args': list_args,
    #                                    'general_param': general_param},
    #                    # -- attachments are handled differently
    #                    'attachments':
    #                        {'time_module': pickle.dumps(time.time)}
                        }
    except Exception as e:
        print("error", str(e))
        returned_dic = {'status': STATUS_FAIL,
                      'loss': 1.0, #debug: probably useless
                      'exception': str(e),
                    # -- store other results like this
                        'eval_time': time.time(),
#                        'other_stuff': {'list_args': list_args,
#                                        'general_param': general_param},
#                        # -- attachments are handled differently
#                        'attachments':
#                            {'time_module': pickle.dumps(time.time)}
                            }

    ### SAVE FILE OF PAST SIMULATION
    try:
        if score_norm is not None:
            json_filename = 'err%.3f_'%score_norm+'hyperopt_results_1call_s'+str(seed)
        else:
            json_filename = 'exception_error__'+'hyperopt_results_1call_s'+str(seed)
        json_dic = {'returned_dic': returned_dic,
                    'current_params': current_params}

        with open(SAVE_DIR+json_filename+'.json','w') as fout:
            # json.dump(json_dic, fout)
            json.dump(json_dic, fout, separators=(',', ':'), sort_keys=True, indent=4)
    except:
        print("WARNING: Results of current simulation were NOT saved correctly to JSON file.")
    # and test if results were saved correctly
    try:
        with open(SAVE_DIR+json_filename+'.json','r') as fin:
            data = json.load(fin)
            #pprint.pprint(byteify(data)) # in order to remove "u" in all keys of dictionary
        print("### Results of current simulation were saved correctly to JSON file. ###")
    except:
        print("WARNING: Results of current simulation were NOT saved correctly to JSON file.")
    ###

    print("returned_dic:", returned_dic)

    return returned_dic


def hp_search(sim_name):
    """
    Main method that is called by the __main__ method when executed.
    It is setting all things necessary for the parameter search and saving the results.
    """

    global SAVE_DIR
    """ Create a unique folder for the experiment with ID value (i.e. time stamp) """
    SAVE_DIR += time.strftime("%Y-%m-%d_%Hh%M__")+sim_name+"_"+str(int(time.time()*10**6))+"/"
    import os
    os.mkdir(SAVE_DIR)
    global sim_name_
    sim_name_ = sim_name

    # Save global parameters in the params variable
    params = {
                'SAVE_DIR': SAVE_DIR,
                'SAVE_TRIAL_OBJ_EVERY': SAVE_TRIAL_OBJ_EVERY,
                'HP_WARMUP_EVALS': HP_WARMUP_EVALS,
                'HP_MAX_EVALS': HP_MAX_EVALS,
                # TODO: INSERT YOUR OWN PARAMETERS BELOW >>>
                'FICHIER_OPTIM2LIVE': sim_name_+FICHIER_OPTIM2LIVE,
                'FICHIER_LIVE2OPTIM': sim_name_+FICHIER_LIVE2OPTIM,
                'DELIMITATEUR': DELIMITATEUR,
                'SCORE_MIN': SCORE_MIN,
                'SCORE_MAX': SCORE_MAX,
                'WAIT_TIME_READ_FILE': WAIT_TIME_READ_FILE
                # <<< INSERT YOUR OWN PARAMETERS HERE
                }

    # save parameters with simulation name before launching simulation
    with open(SAVE_DIR+sim_name_+'_START.json', 'w') as f:
        json.dump(params, f, separators=(',', ':'), sort_keys=True, indent=4)

    global trials
    trials = Trials()

    """ TODO: SOME EXEMPLE OF SEARCH SPACES - MAKE YOUR OWN BELOW >>>"""
    ### NOTE : pour ne pas faire bouger un paramètre il faut faire comme ceci :
    # 'shape': hp.quniform('shape', 0, 0.1, 1),
    search_space = {
                    # LOOK AT THE VARIETY OF KIND OF SPACE PARAMETERS YOU CAN HAVE (uniform, quniform, loguniform, choice, ...)
                    'sr': hp.quniform('sr', 1, 1.1, 1),
                    'iss':hp.loguniform('iss',np.log(10**-2), np.log(5*10**2)),
                    'N':hp.quniform('N', 100, 101, 100),
                    'leak':hp.loguniform('leak', np.log(10**-4), np.log(1.)), # equivalent to hp.uniform('leak', 4.5399929762484854e-05, 1.),
                    'thres_repl_unfrq_w':hp.quniform('thres_repl_unfrq_w', 0, 15,3),
                    'ridge': hp.loguniform('ridge', np.log(10**-11), np.log(1.1*10**1)),
                    'Wstd': hp.loguniform('Wstd', np.log(2*10**-2), np.log(2*10**1)),
                    ## FEEDBACK
                    'fb_func': hp.choice('fb_func', ['heavyside01', 'linear01', 'linear']),
                    'fb_scale': hp.loguniform('fb_scale', np.log(10**-4), np.log(10**1)),
                    'fb_proba': hp.quniform('fb_proba',  0.05, 1.0, 0.05)
                    }
    # search_space = {
    #                  'shape': hp.quniform('shape', 0, 127, 1), #Shape : [0..42] = Sinus / [43..84] = Scie / [85..126] = Carré / 127 = Noise
    #                  'freq-Mod-Env': hp.quniform('freq-Mod-Env', 0, 127, 1), #Freq Mod < Env
    #                  'octave': hp.quniform('octave', 0, 127, 1),
    #                  'VCA-decay': hp.quniform('VCA-decay', 0, 127, 1),
    #                  'VCA-release': hp.quniform('VCA-release', 0, 127, 1),
    #                  'VCF-attack': hp.quniform('VCF-attack', 0, 127, 1),
    #                  'VCF-decay': hp.quniform('VCF-decay', 0, 127, 1),
    #                  'transpose': hp.quniform('transpose', 0, 127, 1) # (-12st..+12st) (j'ai rajouté ça pour changer de note un peu !!!!)
    #                  # volume : non exploré, car réglé par l'utilisateur
    #                 }

    best = fmin(objective_dic,
            space=search_space,
            algo=partial(tpe.suggest, n_startup_jobs=HP_WARMUP_EVALS), #TPE with 'n_startup_jobs' initial random exploration
            # algo=partial(rand.suggest, n_startup_jobs=HP_WARMUP_EVAL), #RANDOM EXPLORATION
            max_evals=HP_MAX_EVALS,
            trials=trials)

    """ Show results with best """
    print("*** trials ***")
    pprint.pprint(trials.trials)
    print("*** trials[0]['result'] ***")
    pprint.pprint(trials.trials[0]['result'])
    print("space_eval"), space_eval(search_space, best)
    print("best"), best


    """ Save results with particular format for hp_analyse_error.py """
    all_loss = [t['result']['loss'] for t in trials.trials]
    all_result = [t['result'] for t in trials.trials]
    all_vals = [t['misc']['vals'] for t in trials.trials]
#    all_res = [t['misc']['vals'].update({'loss':t['result']['loss']}) for t in trials.trials]
    tup_result_vals = [(t['result'], t['misc']['vals']) for t in trials.trials] #for saving useful and seriazable data in a JSON file
    print("all_result")
    pprint.pprint(all_result)
    print("all_values")
    pprint.pprint(all_vals)
    print("max(all_loss)"), max(all_loss)
    print("min(all_loss)"), min(all_loss)

    # save results to JSON file
    json_dic = {'best': best,
                # 'general_param':general_param,
                'params':params,
                'tuples_result_vals': tup_result_vals}
#    with open('hyperopt_results_ter_iss075.json','w') as fout:
    with open(SAVE_DIR+'hyperopt_results_demo.json','w') as fout:
        # json.dump(json_dic, fout)
        json.dump(json_dic, fout, separators=(',', ':'), sort_keys=True, indent=4)
    # and test if results were saved correctly
    try:
#        with open('hyperopt_results_ter_iss075.json','r') as fin:
        with open(SAVE_DIR+'hyperopt_results_demo.json','r') as fin:
            data = json.load(fin)
            # pprint.pprint(byteify(data)) # in order to remove "u" in all keys of dictionary
        print("### Results were saved correctly to JSON file. ###")
    except Exception as e:
        # except:
        print(e)
        print("WARNING: Results were NOT saved correctly to JSON file.")


    save(trials, SAVE_DIR+"hyperopt_trials.pkl")
    save(best, SAVE_DIR+"hyperopt_best.pkl")

    test_open_saved_pickle(SAVE_DIR+"hyperopt_trials.pkl")
    test_open_saved_pickle(SAVE_DIR+"hyperopt_best.pkl")

    # save parameters with simulation name after launching simulation
    # with open(SAVE_DIR+args.filename+'_HP-SEARCH_'+sim_name+'_END.json', 'w') as f:
    with open(SAVE_DIR+sim_name+'_END.json', 'w') as f:
        # json.dump(params,f)
        json.dump(params, f, separators=(',', ':'), sort_keys=True, indent=4)


    #ring bell to say it's finished
    # os.system('play --no-show-progress --null --channels 1 synth %s sine %f' % (1, 1000))


if __name__=='__main__':
    """
    Examples of how to run the hyper-parameter search
    > python hyperapero.py -s vrai
    ou
    > python3 hyperapero.py -s vrai
    """
    # import json
    import argparse

    parser = argparse.ArgumentParser()
    # parser.add_argument("-f", "--filename", type=str, help="json file containing parameters for experiment (root directory is assumed to be params/)")
    parser.add_argument("-s", "--simulation_name", type=str, default="", help="simulation_name: this will be use for the file and saving directory)")
    # parser.add_argument("-l", "--lookup", action='store_true', help="lookup_search_space: Plot the search space given instead of launching a parameter search.")
    args = parser.parse_args()

    # Test to call dummy function to see if read/write files is fine
    # function_that_returns_the_loss_given_the_parameters(params={'shape':0, 'freq-Mod-Env':0})

    # On test si on arrive bien à accéder aux fichiers input/output d'échanges de paramètres ...
    print("\nOn check(e) si tout va bien ...\n")
    test_file_existance(sim_name=args.simulation_name)
    # ... puis on affiche un message pour dire qu'on est prêt
    input("... c'est bon ! On peut commencer l'hyper quest ! (appuie sur une touche s'il te plait !)\n")

    hp_search(sim_name=args.simulation_name)



from typing import Tuple

import numpy as np

from scipy.integrate import solve_ivp


def lorenz(n_timesteps: int, state0: list=[1.0, 1.0, 1.0], 
           rho: float=28.0, sigma: float=10.0, beta: float=8.0/3.0, h: float=0.03) -> Tuple[np.ndarray, np.ndarray]:
    """Lorenz system timeseries.
    .. math::
        \dv{x}{t} = \sigma (y-x)
    .. math::
        \dv{y}{t} = x(\rho - z) - y
    .. math::
        \dv{z}{t} = xy - \beta z

    Args:
        n_timesteps (int): Number of timesteps to generate.
        state0 (list, optional): Initial condition of the system. Defaults to [1.0, 1.0, 1.0].
        rho (float, optional): Rho parameter of the system. Defaults to 28.0.
        sigma (float, optional): Sigma parameter of the system. Defaults to 10.0.
        beta (float, optional): Beta parameter of the system. Defaults to 8.0/3.0.
        h (float, optional): Control the time delta between two timesteps. Defaults to 0.03.

    Returns:
        Tuple[np.ndarray, np.ndarray]: Generated timeseries and timesteps.
    """
    def lorenz_diff(t, state):
        x, y, z = state 
        return sigma * (y - x), x * (rho - z) - y, x * y - beta * z

    t = np.arange(0, n_timesteps * h, h)
    
    sol = solve_ivp(lorenz_diff, 
                    y0=state0, 
                    t_span=(0.0, n_timesteps*h),
                    dense_output=True)
    
    return sol.sol(t).T, t


"""Simple, fast, parallelizable and object-oriented
implementation of Echo State Networks, using offline
learning methods.
"""

# @author: Xavier HINAUT
# xavier.hinaut@inria.fr
# Copyright Xavier Hinaut 2018
# We would like to thank Mantas Lukosevicius for his code that
# was used as inspiration for this code:
# http://minds.jacobs-university.de/mantas/code

import time

from typing import Sequence, Callable, Tuple, Union, Dict
from tempfile import mkdtemp
from pathlib import Path

import joblib
import numpy as np
from scipy import linalg
from tqdm import tqdm

from .utils import check_values, _save
from .regression_models import sklearn_linear_model
from .regression_models import ridge_linear_model
from .regression_models import pseudo_inverse_linear_model


class ESN:
    """Base class of Echo State Networks.

    The :py:class:`.ESN` class is the angular stone of ReservoirPy
    offline learning methods using reservoir computing.
    Echo State Networks allows one to:
        - quickly build ESNs, using the :py:mod:`reservoirpy.mat_gen` module
        to initialize weights,
        - train and test ESNs on the task of your choice,
        - use the trained ESNs on the task of your choice, either in
        predictive mode or generative mode.

    Parameters
    ----------
        lr: float
            Leaking rate
        W: np.ndarray
            Reservoir weights matrix
        Win: np.ndarray
            Input weights matrix
        input_bias: bool, optional
            If True, will add a constant bias
            to the input vector (default: {True})
        reg_model: Callable, optional
            A scikit-learn linear model function to use for
            regression. Should be None if ridge is used. (default: {None})
        ridge: float, optional
            Ridge regularization coefficient for Tikonov regression.
            Should be None if reg_model is used. (default: {None})
        Wfb: np.array, optional
            Feedback weights matrix. (default: {None})
        fbfunc: Callable, optional
            Feedback activation function. (default: {None})
        typefloat: numpy.dtype, optional

    Attributes
    ----------
        Wout: np.ndarray
            Readout matrix
        dim_out: int
            Output dimension
        dim_inp: int
            Input dimension
        N: int
            Number of neuronal units

    See also
    --------
        reservoirpy.ESNOnline for ESN with online learning using FORCE.

    """
    # memmap temporay storage
    # is cleaned after training/running
    _tempstates = Path(mkdtemp(), "states.dat")
    _tempteach = Path(mkdtemp(), "teachers.dat")

    def __init__(self,
                 lr: float,
                 W: np.ndarray,
                 Win: np.ndarray,
                 input_bias: bool = True,
                 reg_model: Callable = None,
                 ridge: float = None,
                 Wfb: np.ndarray = None,
                 fbfunc: Callable = None,
                 typefloat: np.dtype = np.float64):

        self.W = W
        self.Win = Win
        # output weights matrix. must be learnt through training.
        self.Wout = None
        self.Wfb = Wfb

        # check if dimensions of matrices are coherent
        self._autocheck_dimensions()
        self._autocheck_nan()

        # number of neurons
        self.N = self.W.shape[1]
        self.in_bias = input_bias
        # dimension of inputs (including the bias at 1)
        self.dim_inp = self.Win.shape[1]
        self.dim_out = None
        if self.Wfb is not None:
            # dimension of outputs
            self.dim_out = self.Wfb.shape[1]

        self.typefloat = typefloat
        self.lr = lr  # leaking rate

        self.reg_model = self._get_regression_model(ridge, reg_model)
        self.fbfunc = fbfunc
        if self.Wfb is not None and self.fbfunc is None:
            raise ValueError("If a feedback matrix is provided, fbfunc must"
                             f"be a callable object, not {self.fbfunc}.")

    def __repr__(self):
        trained = True
        if self.Wout is None:
            trained = False
        fb = self.Wfb is not None

        out = f"ESN(trained={trained}, feedback={fb}, N={self.N}, "
        out += f"lr={self.lr}, input_bias={self.in_bias}, input_dim={self.N})"
        return out

    # TODO: simplify this with a dict
    def _get_regression_model(self,
                              ridge: float = None,
                              sklearn_model: Callable = None):
        """Set the type of regression used in the model. All regression models available
        for now are described in reservoipy.regression_models:
            - any scikit-learn linear regression model (like Lasso or Ridge)
            - Tikhonov linear regression (l1 regularization)
            - Solving system with pseudo-inverse matrix
        """
        if ridge is not None and sklearn_model is not None:
            raise ValueError("ridge and sklearn_model can not be"
                             "defined at the same time.")

        elif ridge is not None:
            self.ridge = ridge
            return ridge_linear_model(self.ridge)

        elif sklearn_model is not None:
            self.sklearn_model = sklearn_model
            return sklearn_linear_model(self.sklearn_model)

        else:
            return pseudo_inverse_linear_model()

    def _autocheck_nan(self):
        """ Auto-check to see if some important variables do not have
        a problem (e.g. NAN values).
        """
        assert np.isnan(self.Win).any() == False, \
            "Win matrix should not contain NaN values."
        if self.Wfb is not None:
            assert np.isnan(self.Wfb).any() == False, \
                "Wfb matrix should not contain NaN values."

    def _autocheck_dimensions(self):
        # W dimensions check list
        assert len(self.W.shape) == 2, f"W shape should be (N, N) but is {self.W.shape}."
        assert self.W.shape[0] == self.W.shape[1], f"W shape should be (N, N) but is {self.W.shape}."

        # Win dimensions check list
        assert len(self.Win.shape) == 2, f"Win shape should be (N, input) but is {self.Win.shape}."
        err = f"Win shape should be ({self.W.shape[1]}, input) but is {self.Win.shape}."
        assert self.Win.shape[0] == self.W.shape[0], err

    def _autocheck_io(self, inputs, outputs=None):
        # Check if inputs and outputs are lists
        assert type(inputs) is list, "Inputs should be a list of numpy arrays"
        if outputs is not None:
            assert type(outputs) is list, "Outputs should be a list of numpy arrays"

        # check if Win matrix has coherent dimensions with input dimensions
        if self.in_bias:
            err = f"With bias, Win matrix should be of shape ({self.N}, "
            err += f"{inputs[0].shape[1] + 1}) but is {self.Win.shape}."
            assert self.Win.shape[1] == inputs[0].shape[1] + 1, err
        else:
            err = f"Win matrix should be of shape ({self.N}, "
            err += f"{self.dim_inp}) but is {self.Win.shape}."
            assert self.Win.shape[1] == inputs[0].shape[1], err

        if outputs is not None:
            # check feedback matrix
            if self.Wfb is not None:
                err = f"With feedback, Wfb matrix should be of shape ({self.N}, "
                err += f"{outputs[0].shape[1]}) but is {self.Wfb.shape}."
                assert outputs[0].shape[1] == self.Wfb.shape[1], err

    def _get_next_state(self,
                        single_input: np.ndarray,
                        feedback: np.ndarray = None,
                        last_state: np.ndarray = None) -> np.ndarray:
        """Given a state vector x(t) and an input vector u(t),
        compute the state vector x(t+1).

        Parameters
        ----------
            single_input: np.ndarray
                Input vector u(t)

            feedback: numpy.ndarray, optional
                Feedback vector if enabled.
            last_state: numpy.ndarray, optional
                Current state to update x(t). Default to 0 vector.

        Raises
        ------
            RuntimeError: feedback is enabled but no feedback vector is available.

        Returns
        -------
            numpy.ndarray
                Next state x(t+1)
        """

        # check if the user is trying to add empty feedback
        if self.Wfb is not None and feedback is None:
            raise RuntimeError("Missing a feedback vector.")

        # first initialize the current state of the ESN
        if last_state is None:
            x = np.zeros((self.N, 1), dtype=self.typefloat)
        else:
            x = last_state

        # add bias
        if self.in_bias:
            u = np.hstack((1, single_input)).astype(self.typefloat)
        else:
            u = single_input

        # linear transformation
        x1 = np.dot(self.Win, u.reshape(self.dim_inp, 1)) \
            + self.W.dot(x)

        # add feedback if requested
        if self.Wfb is not None:
            x1 += np.dot(self.Wfb, self.fbfunc(feedback))

        # previous states memory leak and non-linear transformation
        x1 = (1-self.lr)*x + self.lr*np.tanh(x1)

        # return the next state computed
        return x1

    def _compute_states(self,
                        input: np.ndarray,
                        forced_teacher: np.ndarray = None,
                        init_state: np.ndarray = None,
                        init_fb: np.ndarray = None,
                        wash_nr_time_step: int = 0,
                        input_id: int = None,
                        memmap: np.memmap = None,
                        input_pos: int = None
                        ) -> Union[Tuple[np.ndarray, np.ndarray], np.ndarray]:
        """Compute all states generated from a single sequence of inputs.

        Parameters
            input {np.ndarray} -- Sequence of inputs.

        Keyword Arguments:
            forced_teacher {np.ndarray} -- Ground truth vectors to use as feedback
                                           during training, if feedback is enabled.
                                           (default: {None})
            init_state {np.ndarray} -- Initialization vector for states. (default: {None})
            init_fb {np.ndarray} -- Initialization vector for feedback. (default: {None})
            wash_nr_time_step {int} -- Number of states to considered as transitory
                            when training. (default: {0})
            input_id {int} -- Index of the input in the queue. Used for parallelization
                              of computations. (default: {None})

        Raises:
            RuntimeError: raised if no teachers are specifiyed for training with feedback.

        Returns:
            Union[Tuple[np.ndarray, np.ndarray], np.ndarray] -- Index of the input in queue
            and computed states, or just states if no index is provided.
        """

        if self.Wfb is not None and forced_teacher is None and self.Wout is None:
            raise RuntimeError("Impossible to use feedback without readout"
                               "matrix or teacher forcing.")

        # to track successives internal states of the reservoir
        states = np.zeros((self.N, len(input)-wash_nr_time_step), dtype=self.typefloat)

        # if a feedback matrix is available, feedback will be set to 0 or to
        # a specific value.
        if self.Wfb is not None:
            if init_fb is None:
                last_feedback = np.zeros((self.dim_out, 1), dtype=self.typefloat)
            else:
                last_feedback = init_fb.copy()
        else:
            last_feedback = None

        # State is initialized to 0 or to a specific value.
        if init_state is None:
            current_state = np.zeros((self.N, 1),dtype=self.typefloat)
        else:
            current_state = init_state.copy().reshape(-1, 1)

        # for each time step in the input
        for t in range(input.shape[0]):
            # compute next state from current state
            current_state = self._get_next_state(input[t, :],
                                                 feedback=last_feedback,
                                                 last_state=current_state)

            # compute last feedback
            if self.Wfb is not None:
                # during training outputs are equal to teachers for feedback
                if forced_teacher is not None:
                    last_feedback = forced_teacher[t,:].reshape(
                        forced_teacher.shape[1], 1).astype(self.typefloat)
                # feedback of outputs, computed with Wout
                else:
                    last_feedback = np.dot(self.Wout,
                                           np.vstack((1, current_state))).astype(self.typefloat)
                last_feedback = last_feedback.reshape(self.dim_out, 1)

            # will track all internal states during inference, and only the
            # states after wash_nr_time_step during training.
            if t >= wash_nr_time_step:
                states[:, t-wash_nr_time_step] = current_state.reshape(-1,).astype(self.typefloat)

        if input_id is None:
            return 0, states

        if memmap is not None:
            memmap[:, input_pos[0]: input_pos[1]] = np.vstack(
                (np.ones((1, states.shape[1]), dtype=self.typefloat),
                 states))

        return input_id, states

    def compute_all_states(self,
                           inputs: Sequence[np.ndarray],
                           forced_teachers: Sequence[np.ndarray] = None,
                           init_state: np.ndarray = None,
                           init_fb: np.ndarray = None,
                           wash_nr_time_step: int = 0,
                           workers: int = -1,
                           backend: str = "threading",
                           verbose: bool = True,
                           memmap: np.memmap = None,
                           ) -> Sequence[np.ndarray]:
        """Compute all states generated from sequences of inputs.

        Parameters
        ----------
            inputs: list or array of numpy.array
                All sequences of inputs used for internal state computation.
                Note that it should always be a list of sequences, i.e. if
                only one sequence of inputs is used, it should be alone in a
                list

            forced_teachers: list or array of numpy.array, optional
                Sequence of ground truths, for computation with feedback without
                any trained readout. Note that is should always be a list of
                sequences of the same length than the `inputs`, i.e. if
                only one sequence of inputs is used, it should be alone in a
                list.

            init_state: np.ndarray, optional
                State initialization vector for all inputs. By default, state
                is initialized at 0.

            init_fb: np.ndarray, optional
                Feedback initialization vector for all inputs, if feedback is
                enabled. By default, feedback is initialized at 0.

            wash_nr_time_step: int, optional
                Number of states to consider as transient when training, and to
                remove when computing the readout weights. By default, no states are
                removed.

            workers: int, optional
                If n >= 1, will enable parallelization of states computation with
                n threads/processes, if possible. If n = -1, will use all available
                resources for parallelization. By default, -1.

            backend: {"threadings", "multiprocessing", "loki"}, optional
                Backend used for parallelization of states computations.
                By default, "threading".

            verbose: bool

        Returns:
            list of np.ndarray
                All computed states.
        """

        # initialization of workers
        loop = joblib.Parallel(n_jobs=workers, backend=backend)
        delayed_states = joblib.delayed(self._compute_states)

        # progress bar if needed
        if verbose:
            track = tqdm
        else:
            def track(x, text):
                return x

        inputs_ends = np.cumsum([i.shape[0] for i in inputs])
        inputs_starts = [end - i.shape[0] for i, end in zip(inputs, inputs_ends)]

        # no feedback training or running
        if forced_teachers is None:
            all_states = loop(delayed_states(inputs[i],
                                             wash_nr_time_step=wash_nr_time_step,
                                             input_id=i,
                                             init_state=init_state,
                                             init_fb=init_fb,
                                             memmap=memmap,
                                             input_pos=(inputs_starts[i], inputs_ends[i]))
                              for i in track(range(len(inputs)), "Computing states"))
        # feedback training
        else:
            all_states = loop(delayed_states(inputs[i],
                                             forced_teachers[i],
                                             wash_nr_time_step=wash_nr_time_step,
                                             input_id=i,
                                             init_state=init_state,
                                             init_fb=init_fb,
                                             memmap=memmap,
                                             input_pos=(inputs_starts[i], inputs_ends[i]))
                              for i in track(range(len(inputs)), "Computing states"))

        # input ids are used to make sure that the returned states are in the same order
        # as inputs, because parallelization can change this order.
        return [s[1] for s in sorted(all_states, key=lambda x: x[0])]

    def compute_outputs(self,
                        states: Sequence[np.ndarray],
                        verbose: bool = False
                        ) -> Sequence[np.ndarray]:
        """Compute all readouts of a given sequence of states,
        when a readout matrix is available (i.e. after training).

        Parameters
        ----------
            states: list of numpy.array
                All sequences of states used for readout.

            verbose: bool

        Raises
        ------
            RuntimeError: no readout matrix Wout is available.
            Consider training model first, or load an existing matrix.

        Returns
        -------
            list of numpy.arrays
                All outputs of readout matrix.
        """
        # because all states and readouts will be concatenated,
        # first save the indexes of each inputs states in the concatenated vector.
        if self.Wout is not None:

            if verbose:
                print("Computing outputs...")
                tic = time.time()

            outputs = [None] * len(states)
            for i, s in enumerate(states):
                x = np.vstack((np.ones((s.shape[1],), dtype=self.typefloat), s))
                y = np.dot(self.Wout, x).astype(self.typefloat)
                outputs[i] = y

            if verbose:
                toc = time.time()
                print(f"Outputs computed! (in {toc - tic}sec)")

            return outputs

        else:
            raise RuntimeError("Impossible to compute outputs: no readout matrix available.")

    def fit_readout(self,
                    states: Sequence,
                    teachers: Sequence,
                    reg_model: Callable = None,
                    ridge: float = None,
                    force_pinv: bool = False,
                    verbose: bool = False,
                    use_memmap: bool = False) -> np.ndarray:
        """Compute a readout matrix by fitting the states computed by the ESN
        to the expected values, using the regression model defined
        in the ESN.

        Parameters
        ----------
            states: list of numpy.ndarray
                All states computed.

            teachers: list of numpy.ndarray
                All ground truth vectors.

            reg_model: scikit-learn regression model, optional
                A scikit-learn regression model to use for readout
                weights computation.

            ridge: float, optional
                Use Tikhonov regression for readout weights computation
                and set regularization parameter to the parameter value.

            force_pinv: bool, optional
                Overwrite all previous parameters and
                force computation of readout using pseudo-inversion.

            verbose: bool

        Returns
        -------
            numpy.ndarray
                Readout matrix.
        """
        # switch the regression model used at instanciation if needed.
        # WARNING: this change won't be saved by the save function.
        if (ridge is not None) or (reg_model is not None):
            reg_model = self._get_regression_model(ridge, reg_model)
        elif force_pinv:
            reg_model = self._get_regression_model(None, None)
        else:
            reg_model = self.reg_model

        # check if network responses are valid
        check_values(array_or_list=states, value=None)

        if verbose:
            tic = time.time()
            print("Linear regression...")
        # concatenate the lists (along timestep axis)
        if not use_memmap:
            X = np.hstack(states).astype(self.typefloat)
            Y = np.hstack(teachers).astype(self.typefloat)

            # Adding ones for regression with bias b in (y = a*x + b)
            X = np.vstack((np.ones((1, X.shape[1]), dtype=self.typefloat), X))

            # Building Wout with a linear regression model.
            # saving the output matrix in the ESN object for later use
            Wout = reg_model(X, Y)

        else:
            Wout = reg_model(states, teachers)
            del states
            del teachers

        if verbose:
            toc = time.time()
            print(f"Linear regression done! (in {toc - tic} sec)")

        # return readout matrix
        return Wout

    def train(self,
              inputs: Sequence[np.ndarray],
              teachers: Sequence[np.ndarray],
              wash_nr_time_step: int = 0,
              workers: int = -1,
              backend: str = "threading",
              verbose: bool = False,
              use_memmap: bool = False) -> Sequence[np.ndarray]:
        """Train the ESN model on set of input sequences.

        Arguments
        ---------
            inputs: list of numpy.ndarray
                List of inputs.
                Note that it should always be a list of sequences, i.e. if
                only one sequence of inputs is used, it should be alone in a
                list.

            teachers: list of numpy.ndarray
                List of ground truths.
                Note that is should always be a list of
                sequences of the same length than the `inputs`, i.e. if
                only one sequence of inputs is used, it should be alone in a
                list.

            wash_nr_time_step: int
                Number of states to considered as transient when training. Transient
                states will be discarded when computing readout matrix. By default,
                no states are removes.

            workers: int, optional
                If n >= 1, will enable parallelization of states computation with
                n threads/processes, if possible. If n = -1, will use all available
                resources for parallelization. By default, -1.

            backend: {"threadings", "multiprocessing", "loki"}, optional
                Backend used for parallelization of states computations.
                By default, "threading".

            verbose: bool

        Returns
        -------
            list of numpy.ndarray
                All states computed, for all inputs.

        Note
        ----
            If only one input sequence is provided ("continuous time" inputs), workers should be 1,
            because parallelization is impossible. In other cases, if using large NumPy arrays during
            computation (which is often the case), prefer using `threading` backend to avoid huge
            overhead. Multiprocess is a good idea only in very specific cases, and this code is not
            (yet) well suited for this.
        """
        ## Autochecks of inputs and outputs
        self._autocheck_io(inputs=inputs, outputs=teachers)

        steps = np.sum([i.shape[0] for i in inputs])
        if verbose:
            print(f"Training on {len(inputs)} inputs ({steps} steps)-- wash: {wash_nr_time_step} steps")

        memstates = None
        if use_memmap:
            memstates = np.memmap(self._tempstates, dtype=self.typefloat, mode="w+",
                                  shape=(self.N + 1, steps))

        # compute all states
        all_states = self.compute_all_states(inputs,
                                             forced_teachers=teachers,
                                             wash_nr_time_step=wash_nr_time_step,
                                             workers=workers,
                                             backend=backend,
                                             verbose=verbose,
                                             memmap=memstates)

        all_teachers = [t[wash_nr_time_step:].T for t in teachers]

        # compute readout matrix
        if use_memmap:
            memteachers = np.memmap(self._tempteach, dtype=self.typefloat, mode="w+",
                                    shape=(all_teachers[0].shape[0], steps))
            memteachers[:] = np.hstack(all_teachers)
            self.Wout = self.fit_readout(memstates, memteachers, use_memmap=True, verbose=verbose)
        else:
            self.Wout = self.fit_readout(all_states, all_teachers, verbose=verbose)

        # save the expected dimension of outputs
        self.dim_out = self.Wout.shape[0]

        # return all internal states
        return [st.T for st in all_states]

    def run(self,
            inputs: Sequence[np.ndarray],
            init_state: np.ndarray=None,
            init_fb: np.ndarray=None,
            workers: int=-1,
            backend: str="threading",
            verbose: bool=False) -> Tuple[Sequence[np.ndarray], Sequence[np.ndarray]]:
        """Run the model on a sequence of inputs, and returned the states and
           readouts vectors.

        Arguments:
            inputs {Sequence[np.ndarray]} -- Sequence of inputs.

        Keyword Arguments:
            init_state {np.ndarray} -- State initialization vector
                                       for all inputs. (default: {None})
            init_fb {np.ndarray} -- Feedback initialization vector
                                    for all inputs, if feedback is
                                    enabled. (default: {None})
            workers {int} -- if n >= 1, will enable parallelization of
                             states computation with n threads/processes, if possible.
                             if n = -1, will use all available resources for
                             parallelization.
            backend {str} -- Backend used for parallelization of
                             states computations. Available backends are
                             `threadings`(recommended, see Note), `multiprocess`,
                             `loky` (default: {"threading"}).
            verbose {bool} -- if `True`, display progress in stdout.

        Returns:
            Tuple[Sequence[np.ndarray], Sequence[np.ndarray]] -- All states and readouts,
                                                                 for all inputs.

        Note:
            If only one input sequence is provided ("continuous time" inputs), workers should be 1,
            because parallelization is impossible. In other cases, if using large NumPy arrays during
            computation (which is often the case), prefer using `threading` backend to avoid huge
            overhead. Multiprocess is a good idea only in very specific cases, and this code is not
            (yet) well suited for this.
        """

        if verbose:
            steps = np.sum([i.shape[0] for i in inputs])
            print(f"Running on {len(inputs)} inputs ({steps} steps)")

        ## Autochecks of inputs
        self._autocheck_io(inputs=inputs)

        all_states = self.compute_all_states(inputs,
                                             init_state=init_state,
                                             init_fb=init_fb,
                                             workers=workers,
                                             backend=backend,
                                             verbose=verbose)

        all_outputs = self.compute_outputs(all_states)
        # return all_outputs, all_int_states
        return [st.T for st in all_outputs], [st.T for st in all_states]

    def generate(self,
                 nb_timesteps: int,
                 init_inputs: np.ndarray,
                 init_state: np.ndarray = None,
                 init_fb: np.ndarray = None,
                 return_init: bool = False,
                 verbose: bool = False
                 ) -> np.ndarray:
        """Run the ESN on a generative mode.
        After the init_inputs are consumed, new outputs are
        used as inputs for the next nb_timesteps.

        Args:
            nb_timesteps (int): [description]
            init_inputs (np.ndarray): [description]
            init_state (np.ndarray, optional): [description]. Defaults to None.
            init_fb (np.ndarray, optional): [description]. Defaults to None.
            verbose (bool, optional): [description]. Defaults to False.

        Returns:
            np.ndarray: [description]
        """

        if verbose:
            print(f"Generating {nb_timesteps} timesteps from {init_inputs.shape[0]} inputs.")
            print("Computing initial states...")

        _, init_states = self._compute_states(init_inputs, init_state=init_state,
                                              init_fb=init_fb)

        s0 = init_states[:, -1].reshape(-1, 1)
        init_outputs = self.compute_outputs([init_states])[0]
        u0 = init_outputs[:, -1].reshape(1, -1)

        if init_fb is not None:
            fb0 = self.compute_outputs([init_states[:, -2]])[0]
        else:
            fb0 = None

        if verbose:
            track = tqdm
        else:
            track = lambda x, text: x

        states = [None] * nb_timesteps
        outputs = [None] * nb_timesteps
        for i in track(range(nb_timesteps), "Generating"):
            _, s = self._compute_states(u0, init_state=s0, init_fb=fb0)
            s0 = s[:, -1].reshape(-1, 1)
            states[i] = s0.reshape(self.N)

            if fb0 is not None:
                fb0 = u0.copy()

            u = self.compute_outputs([s0])
            u0 = u[0].reshape(1, -1)
            outputs[i] = u0.reshape(self.dim_inp - self.in_bias)

        outputs = np.array(outputs)
        states = np.array(states)

        if return_init:
            outputs = np.vstack([init_outputs.T, outputs])
            states = np.vstack([init_states.T, states])

        return outputs, states

    def save(self, directory: str):
        """Save the ESN to disk.

        Arguments:
            directory {str or Path} -- Directory of the saved model.
        """
        _save(self, directory)

    #? maybe remove this in the future: unused
    def describe(self) -> Dict:
        """
        Provide descriptive stats about ESN matrices.

        Returns:
            Dict -- Descriptive data.
        """

        desc = {
            "Win": {
                "max": np.max(self.Win),
                "min": np.min(self.Win),
                "mean": np.mean(self.Win),
                "median": np.median(self.Win),
                "std": np.std(self.Win)
            },
            "W": {
                "max": np.max(self.W),
                "min": np.min(self.W),
                "mean": np.mean(self.W),
                "median": np.median(self.W),
                "std": np.std(self.W),
                "sr": max(abs(linalg.eig(self.W)[0]))
            }
        }
        if self.Wfb is not None:
            desc["Wfb"] = {
                "max": np.max(self.Wfb),
                "min": np.min(self.Wfb),
                "mean": np.mean(self.Wfb),
                "median": np.median(self.Wfb),
                "std": np.std(self.Wfb)
            }
        if self.Wout is not None:
            desc["Wout"] = {
                "max": np.max(self.Wout),
                "min": np.min(self.Wout),
                "mean": np.mean(self.Wout),
                "median": np.median(self.Wout),
                "std": np.std(self.Wout)
            }
        return desc



import pytest
import numpy as np

from numpy.testing import assert_almost_equal
from numpy.testing import assert_array_almost_equal

from reservoirpy.activationsfunc import get_function

from reservoirpy.activationsfunc import softmax
from reservoirpy.activationsfunc import softplus
from reservoirpy.activationsfunc import identity
from reservoirpy.activationsfunc import tanh
from reservoirpy.activationsfunc import sigmoid
from reservoirpy.activationsfunc import relu


@pytest.mark.parametrize("value, expected", [
    ([1, 2, 3], np.exp([1, 2, 3])/ np.sum(np.exp([1, 2, 3]))),
    (1, np.exp(1) / np.sum(np.exp(1))),
    ([0, 0], [0.5, 0.5])
])
def test_softmax(value, expected):
    result = softmax(value)

    assert_almost_equal(np.sum(result), 1.0)
    assert_array_almost_equal(result, expected)


@pytest.mark.parametrize("value, expected", [
    (0, np.log(1 + np.exp(0))),
    ([0, 1, 2], np.log(1 + np.exp([0, 1, 2]))),
    ([-2, -1], np.log(1 + np.exp([-2, -1])))
])
def test_softplus(value, expected):
    result = softplus(value)

    assert np.any(result > 0.0)
    assert_array_almost_equal(result, expected)


@pytest.mark.parametrize("value",[
    ([1, 2, 3]),
    ([1]),
    (0),
    ([0.213565165, 0.1, 1.064598495615132]),
])
def test_identity(value):
    result = identity(value)
    val = np.asanyarray(value)

    assert np.any(result == val)


@pytest.mark.parametrize("value, expected", [
    ([1, 2, 3], np.tanh([1, 2, 3])),
    (0, np.tanh(0))
])
def test_tanh(value, expected):
    result = tanh(value)

    assert_array_almost_equal(result, expected)


@pytest.mark.parametrize("value, expected", [
    ([1, 2, 3], 1 / (1 + np.exp(-np.array([1, 2, 3])))),
    (0, 1 / (1 + np.exp(0))),
    ([-1000, -2], [0.0, 1.35e-1])
])
def test_sigmoid(value, expected):
    result = sigmoid(value)
    assert_array_almost_equal(result, expected, decimal=1)


@pytest.mark.parametrize("value, expected", [
    ([1, 2, 3], np.array([1, 2, 3])),
    ([-1, -10, 5], np.array([0, 0, 5])),
    (0, np.array(0))
])
def test_relu(value, expected):
    result = relu(value)
    assert_array_almost_equal(result, expected)